import os
import time
import warnings

import numpy as np

from pydfc import MultiAnalysis, data_loader

warnings.simplefilter("ignore")

os.environ["MKL_NUM_THREADS"] = "16"
os.environ["NUMEXPR_NUM_THREADS"] = "16"
os.environ["OMP_NUM_THREADS"] = "16"

################################# Parameters #################################
# data paths
# main_root = '../../DATA/ds002785/' # for local
main_root = "../../../DATA/task-based/openneuro/ds002785"  # for server
roi_root = f"{main_root}/derivatives/ROI_timeseries"
output_root = f"{main_root}/derivatives/fitted_MEASURES"

# for consistency we use 0 for resting state
TASKS = [
    "task-restingstate",
    "task-anticipation",
    "task-emomatching",
    "task-faces",
    "task-gstroop",
    "task-workingmemory",
]

job_id = int(os.getenv("SGE_TASK_ID"))
TASK_id = job_id - 1  # SGE_TASK_ID starts from 1 not 0
if TASK_id >= len(TASKS):
    print("TASK_id out of TASKS")
    exit()
task = TASKS[TASK_id]

###### MEASUREMENT PARAMETERS ######

# W is in sec

params_methods = {
    # Sliding Parameters
    "W": 44,
    "n_overlap": 1.0,
    "sw_method": "pear_corr",
    "tapered_window": True,
    # TIME_FREQ
    "TF_method": "WTC",
    # CLUSTERING AND DHMM
    "clstr_base_measure": "SlidingWindow",
    # HMM
    "hmm_iter": 20,
    "dhmm_obs_state_ratio": 16 / 24,
    # State Parameters
    "n_states": 12,
    "n_subj_clstrs": 20,
    # Parallelization Parameters
    "n_jobs": 2,
    "verbose": 0,
    "backend": "loky",
    # SESSION
    "session": task,
    # Hyper Parameters
    "normalization": True,
    "num_subj": None,  # None or 216?
    "num_time_point": None,  # None or set?
}

###### HYPER PARAMETERS ALTERNATIVE ######

MEASURES_name_lst = [
    "SlidingWindow",
    "Time-Freq",
    "CAP",
    "ContinuousHMM",
    "Windowless",
    "Clustering",
    "DiscreteHMM",
]

alter_hparams = {
    # 'session': ['Rest1_RL', 'Rest2_LR', 'Rest2_RL'],
    # 'n_overlap': [0, 0.25, 0.75, 1],
    # 'n_states': [6, 16],
    # # 'normalization': [],
    # 'num_subj': [50, 100, 200],
    # 'num_select_nodes': [30, 50, 333],
    # 'num_time_point': [800, 1000],
    # 'Fs_ratio': [0.50, 0.75, 1.5],
    # 'noise_ratio': [1.00, 2.00, 3.00],
    # 'num_realization': []
}

###### MultiAnalysis PARAMETERS ######

params_multi_analysis = {
    # Parallelization Parameters
    "n_jobs": None,
    "verbose": 0,
    "backend": "loky",
}

################################# LOAD DATA #################################

BOLD = data_loader.load_TS(
    data_root=roi_root, file_name="time_series.npy", SESSIONs=task, subj_id2load=None
)

################################# Visualize BOLD #################################

# for session in BOLD:
#     BOLD.visualize(start_time=0, end_time=2000, nodes_lst=list(range(10)),
#         save_image=False, output_root=None)

################################ Measures of dFC #################################

MA = MultiAnalysis(
    analysis_name=f"task-based-dFC-ds002785-{task}", **params_multi_analysis
)

MEASURES_lst = MA.measures_initializer(MEASURES_name_lst, params_methods, alter_hparams)

tic = time.time()
print("Measurement Started ...")

################################# estimate FCS #################################

for MEASURE_id, measure in enumerate(MEASURES_lst):

    print("MEASURE: " + measure.measure_name)
    print("FCS estimation started...")

    if measure.is_state_based:
        measure.estimate_FCS(time_series=BOLD)

    # dFC_analyzer.estimate_group_FCS(time_series_dict=BOLD)
    print("FCS estimation done.")

    # Save
    if not os.path.exists(f"{output_root}/{task}"):
        os.makedirs(f"{output_root}/{task}")
    np.save(f"{output_root}/{task}/MEASURE_{str(MEASURE_id)}.npy", measure)

print(f"Measurement required {time.time() - tic:0.3f} seconds.")
np.save(f"{output_root}/{task}/multi_analysis.npy", MA)

#################################################################################



"""The pydFC toolbox.

Submodules
---------

dfc_methods                --- implementation of dFC methods
multi_analysis             --- multi analysis class implementing
                               multiple dFC methods simultaneously
time_series                --- time series class
dfc                        --- dfc class
data_loader                --- load data
multi_analysis_utils       --- utility functions for multi analysis
                               to implement multiple dFC methods
                               simultaneously
dfc_utils                  --- functions used for dFC analysis
comparison                 --- functions used for dFC results comparison

"""

from . import dfc_methods
from .dfc import DFC
from .multi_analysis import MultiAnalysis
from .time_series import TIME_SERIES

__all__ = [
    "MultiAnalysis",
    "TIME_SERIES",
    "DFC",
    "data_loader",
    "multi_analysis_utils",
    "dfc_methods",
    "dfc_utils",
    "comparison",
]



# -*- coding: utf-8 -*-
"""
SimilarityAssessment class
functions to assess similarity between dFC results

Created on Jun 29 2023
@author: Mohammad Torabi
"""

from copy import deepcopy

import numpy as np
from joblib import Parallel, delayed
from scipy import stats

from ..dfc_utils import (
    TR_intersection,
    calc_graph_propoerty,
    dFC_mat2vec,
    filter_dFC_lst,
    find_new_order,
    mutual_information,
    normalized_euc_dist,
)

################################# SimilarityAssessment class ####################################


class SimilarityAssessment:

    def __init__(self, dFC_lst):
        self.dFC_lst = dFC_lst

    ##################### dFC FEATURES ######################

    def FO_calc(self, dFC_lst, common_TRs=None):

        # returns, for each state the Fractional Occupancy (FO)
        # see Vidaurre et al., 2017
        # it only considers TRs in common_TRs

        if common_TRs is None:
            common_TRs = TR_intersection(dFC_lst)

        FO_list = list()
        for dFC in dFC_lst:

            FO = {}

            if dFC.measure.is_state_based:

                state_act_dict = dFC.state_act_dict(TRs=common_TRs)

                for FCS_key in state_act_dict["state_TC"]:
                    FO[FCS_key] = np.mean(state_act_dict["state_TC"][FCS_key]["act_TC"])

            FO_list.append(FO)

        return FO_list

    def transition_stats(self, dFC_lst, common_TRs=None):
        # returns the number of total state transition within common_TRs -> trans_freq
        # and the number of total state transitions regardless of common_TRs
        # but normalized by total number of TRs -> trans_norm
        # and a list of all dwell times

        if common_TRs is None:
            common_TRs = TR_intersection(dFC_lst)

        TRs_lst = list()
        for TR in common_TRs:
            TRs_lst.append("TR" + str(TR))

        output_lst = list()
        for dFC in dFC_lst:

            output_dict = {}

            if dFC.measure.is_state_based:

                #  downsampled
                trans_freq = 0
                dwell_time_lst = list()
                dwell_time = 0
                last_TR = None
                for TR in dFC.FCS_idx:
                    if TR in TRs_lst:
                        if not last_TR is None:
                            if dFC.FCS_idx[TR] != dFC.FCS_idx[last_TR]:
                                dwell_time_lst.append(dwell_time)
                                dwell_time = 0
                                trans_freq += 1
                        dwell_time += 1
                        last_TR = TR

                output_dict["dwell_time"] = dwell_time_lst
                output_dict["trans_freq"] = trans_freq

                # normalized (not downsampled)
                trans_norm = 0
                dwell_time_lst = list()
                dwell_time = 0
                last_TR = None
                for TR in dFC.FCS_idx:
                    if not last_TR is None:
                        if dFC.FCS_idx[TR] != dFC.FCS_idx[last_TR]:
                            dwell_time_lst.append(dwell_time / len(dFC.FCS_idx))
                            dwell_time = 0
                            trans_norm += 1
                    dwell_time += 1
                    last_TR = TR
                trans_norm = trans_norm / len(dFC.FCS_idx)

                output_dict["dwell_time_norm"] = dwell_time_lst
                output_dict["trans_norm"] = trans_norm

            output_lst.append(output_dict)

        return output_lst

    def feature_all(self, dFC_mat):
        vectorized_dFC = dFC_mat2vec(dFC_mat).flatten()  # (time*connection, )
        vectorized_dFC = np.expand_dims(vectorized_dFC, axis=0)  # (1, time*connection)
        return vectorized_dFC

    def feature_spatial(self, dFC_mat):
        conn_over_time = dFC_mat2vec(dFC_mat)  # (time, connection)
        return conn_over_time

    def feature_temporal(self, dFC_mat):
        conn_over_time = dFC_mat2vec(dFC_mat)  # (time, connection)
        time_over_conn = conn_over_time.T  # (connection, time)
        return time_over_conn

    def feature_inter_time_corr(self, dFC_mat):
        """
        returns correspondence of inter-time relation between results of dFC
        measures in each subject
        """
        conn_over_time = dFC_mat2vec(dFC_mat)  # (time, connection)
        inter_time_corr = np.corrcoef(conn_over_time)  # (time, time)
        inter_time_corr = np.nan_to_num(inter_time_corr)
        inter_time_corr = dFC_mat2vec(inter_time_corr)  # (time*(time-1)/2, )
        inter_time_corr = np.expand_dims(inter_time_corr, axis=0)  # (1, time*(time-1)/2)
        return inter_time_corr

    def feature_inter_conn_corr(self, dFC_mat):
        conn_over_time = dFC_mat2vec(dFC_mat)  # (time, connection)
        time_over_conn = conn_over_time.T  # (connection, time)
        inter_conn_corr = np.corrcoef(time_over_conn)  # (connection, connection)
        inter_conn_corr = np.nan_to_num(inter_conn_corr)
        inter_conn_corr = dFC_mat2vec(inter_conn_corr)  # (connection*(connection-1)/2, )
        inter_conn_corr = np.expand_dims(
            inter_conn_corr, axis=0
        )  # (1, connection*(connection-1)/2)
        return inter_conn_corr

    def feature_dFC_avg(self, dFC_mat):
        dFC_avg = np.mean(dFC_mat, axis=0)  # (ROI, ROI)
        vectorized_dFC_avg = dFC_mat2vec(dFC_avg)  # (connection, )
        vectorized_dFC_avg = np.expand_dims(vectorized_dFC_avg, axis=0)  # (1, connection)
        return vectorized_dFC_avg

    def feature_dFC_var(self, dFC_mat):
        dFC_var = np.var(dFC_mat, axis=0)  # (ROI, ROI)
        vectorized_dFC_var = dFC_mat2vec(dFC_var)  # (connection, )
        vectorized_dFC_var = np.expand_dims(vectorized_dFC_var, axis=0)  # (1, connection)
        return vectorized_dFC_var

    def feature_graph_spatial(self, dFC_mat, graph_property):
        graph_feature_over_time = list()
        for FC_mat in dFC_mat:
            graph_feature = calc_graph_propoerty(
                FC_mat, property=graph_property, threshold=False, binarize=False
            )
            graph_feature_over_time.append(graph_feature)
        graph_feature_over_time = np.array(graph_feature_over_time)  # (time, ROI)
        return graph_feature_over_time

    def feature_graph_temporal(self, dFC_mat, graph_property):
        graph_feature_over_time = list()
        for FC_mat in dFC_mat:
            graph_feature = calc_graph_propoerty(
                FC_mat, property=graph_property, threshold=False, binarize=False
            )
            graph_feature_over_time.append(graph_feature)
        graph_feature_over_time = np.array(graph_feature_over_time)  # (time, ROI)
        graph_feature_over_node = graph_feature_over_time.T  # (ROI, time)
        graph_feature_avg = np.mean(graph_feature_over_node, axis=0)  # (time, )
        graph_feature_avg = np.expand_dims(graph_feature_avg, axis=0)  # (1, time)
        return graph_feature_avg

    def extract_feature(self, dFC_mat, feature2extract, graph_property=None):
        """
        feature2extract_list = [
            'all',
            'spatial', 'temporal',
            'inter_time_corr', 'inter_conn_corr',
            'dFC_avg', 'dFC_var',
            'graph_spatial', 'graph_temporal'
        ]
        """
        feature = None
        if feature2extract == "all":
            feature = self.feature_all(dFC_mat)
        if feature2extract == "spatial":
            feature = self.feature_spatial(dFC_mat)
        if feature2extract == "temporal":
            feature = self.feature_temporal(dFC_mat)
        if feature2extract == "inter_time_corr":
            feature = self.feature_inter_time_corr(dFC_mat)
        if feature2extract == "inter_conn_corr":
            feature = self.feature_inter_conn_corr(dFC_mat)
        if feature2extract == "dFC_avg":
            feature = self.feature_dFC_avg(dFC_mat)
        if feature2extract == "dFC_var":
            feature = self.feature_dFC_var(dFC_mat)
        if feature2extract == "graph_spatial":
            feature = self.feature_graph_spatial(dFC_mat, graph_property=graph_property)
        if feature2extract == "graph_temporal":
            feature = self.feature_graph_temporal(dFC_mat, graph_property=graph_property)

        return feature

    def dFC_mat_lst_similarity(
        self, dFC_mat_lst, feature2extract, metric, graph_property=None
    ):

        sim_mat_over_sample = None
        for i, dFC_mat_i in enumerate(dFC_mat_lst):
            for j, dFC_mat_j in enumerate(dFC_mat_lst):

                if j <= i:
                    continue

                assert dFC_mat_i.shape == dFC_mat_j.shape, "shape mismatch"

                feature_i = self.extract_feature(
                    dFC_mat_i,
                    feature2extract=feature2extract,
                    graph_property=graph_property,
                )  # (samples, variables)
                feature_j = self.extract_feature(
                    dFC_mat_j,
                    feature2extract=feature2extract,
                    graph_property=graph_property,
                )  # (samples, variables)

                sim_over_sample = list()
                for sample in range(feature_i.shape[0]):
                    if (
                        np.var(feature_i[sample, :]) == 0
                        or np.var(feature_j[sample, :]) == 0
                    ):
                        sim = 0
                    else:
                        if metric == "corr":
                            sim = np.corrcoef(feature_i[sample, :], feature_j[sample, :])[
                                0, 1
                            ]
                        elif metric == "spearman":
                            sim, p = stats.spearmanr(
                                feature_i[sample, :], feature_j[sample, :]
                            )
                        elif metric == "MI":
                            sim = mutual_information(
                                X=feature_i[sample, :], Y=feature_j[sample, :], N_bins=100
                            )
                        elif metric == "euclidean_distance":
                            # normalized euclidean is used
                            sim = normalized_euc_dist(
                                x=feature_i[sample, :], y=feature_j[sample, :]
                            )
                    sim_over_sample.append(sim)

                if sim_mat_over_sample is None:
                    sim_mat_over_sample = np.zeros(
                        (len(sim_over_sample), len(dFC_mat_lst), len(dFC_mat_lst))
                    )
                sim_mat_over_sample[:, i, j] = np.array(sim_over_sample)
                sim_mat_over_sample[:, j, i] = sim_mat_over_sample[:, i, j]

        return sim_mat_over_sample

    def assess_similarity(self, dFC_lst, downsampling_method="default", **param_dict):
        """
        downsampling_method: 'default' picks FCs at common_TRs
        while 'SWed' uses a sliding window to downsample
        """
        methods_assess = {}

        # sort dFC_lst according to methods names
        old_list = [dFC.measure.measure_name for dFC in dFC_lst]
        new_list = deepcopy(old_list)
        new_list.sort()

        new_order = find_new_order(old_list, new_list)
        dFC_lst = [dFC_lst[i] for i in new_order]

        common_TRs = TR_intersection(dFC_lst)

        measure_lst = list()
        TS_info_lst = list()
        dFC_mat_lst = list()
        for dFC in dFC_lst:
            measure_lst.append(dFC.measure)
            TS_info_lst.append(dFC.TS_info)
            if downsampling_method == "SWed":
                dFC_mat_lst.append(
                    dFC.SWed_dFC_mat(
                        W=param_dict["W"],
                        n_overlap=param_dict["n_overlap"],
                        tapered_window=param_dict["tapered_window"],
                    )
                )
            else:
                dFC_mat_lst.append(dFC.get_dFC_mat(TRs=common_TRs))

        methods_assess["measure_lst"] = measure_lst
        methods_assess["TS_info_lst"] = TS_info_lst
        methods_assess["common_TRs"] = common_TRs

        ########## dFC samples ##########

        dFC_samples = {}
        for i, dFC_mat in enumerate(dFC_mat_lst):
            dFC_samples[str(i)] = dFC_mat
        methods_assess["dFC_samples"] = dFC_samples

        ########## time record ##########

        time_record_dict = {}
        for i, dFC in enumerate(dFC_lst):
            time_record = {}
            time_record["FCS_fit"] = dFC.measure.FCS_fit_time
            time_record["dFC_assess"] = dFC.measure.dFC_assess_time
            time_record_dict[str(i)] = time_record
        methods_assess["time_record_dict"] = time_record_dict

        ########## subj_dFC_sim ##########
        # returns correlation/MI/spearman corr/euclidean distance between results of dFC
        # measures in a subject
        feature2extract_list = [
            # 'all',
            "spatial",
            "temporal",
            "inter_time_corr",
            "inter_conn_corr",
            "dFC_avg",
            "dFC_var",
            # 'graph_spatial', 'graph_temporal'
        ]
        metric_list = ["corr", "spearman", "MI", "euclidean_distance"]
        graph_property_list = ["ECM", "shortest_path", "degree", "clustering_coef"]
        methods_assess["all"] = {}
        for metric in metric_list:
            methods_assess["all"][metric] = self.dFC_mat_lst_similarity(
                dFC_mat_lst, feature2extract="all", metric=metric
            )
        methods_assess["feature_based"] = {}
        for feature2extract in feature2extract_list:
            methods_assess["feature_based"][feature2extract] = (
                self.dFC_mat_lst_similarity(
                    dFC_mat_lst, feature2extract=feature2extract, metric="spearman"
                )
            )
        methods_assess["graph_based"] = {}
        methods_assess["graph_based"]["graph_spatial"] = {}
        methods_assess["graph_based"]["graph_temporal"] = {}
        for graph_property in graph_property_list:
            methods_assess["graph_based"]["graph_spatial"][graph_property] = (
                self.dFC_mat_lst_similarity(
                    dFC_mat_lst,
                    feature2extract="graph_spatial",
                    metric="spearman",
                    graph_property=graph_property,
                )
            )
            methods_assess["graph_based"]["graph_temporal"][graph_property] = (
                self.dFC_mat_lst_similarity(
                    dFC_mat_lst,
                    feature2extract="graph_temporal",
                    metric="spearman",
                    graph_property=graph_property,
                )
            )
        # ########## dFC temporal average and variance ##########

        methods_assess["dFC_avg"] = [
            self.feature_dFC_avg(dFC_mat) for dFC_mat in dFC_mat_lst
        ]

        methods_assess["dFC_var"] = [
            self.feature_dFC_var(dFC_mat) for dFC_mat in dFC_mat_lst
        ]

        ########## Fractional Occupancy ##########

        FO_lst = self.FO_calc(dFC_lst, common_TRs=common_TRs)
        methods_assess["FO"] = FO_lst

        ########## transition frequency ##########

        transition_stats_lst = self.transition_stats(dFC_lst, common_TRs=common_TRs)
        methods_assess["transition_stats"] = transition_stats_lst

        ##############################################
        return methods_assess

    def run(self, FILTERS, downsampling_method="default"):
        """
        downsampling_method: 'default' picks FCs at common_TRs
        while 'SWed' uses a sliding window to downsample
        """
        parallelize = True
        output = {}
        if parallelize:
            out_lst = Parallel(n_jobs=4, verbose=0, backend="loky")(
                delayed(self.assess_similarity)(
                    dFC_lst=filter_dFC_lst(self.dFC_lst, **FILTERS[filter]),
                    downsampling_method=downsampling_method,
                    **FILTERS[filter],
                )
                for filter in FILTERS
            )
            for i, filter in enumerate(FILTERS):
                output[filter] = out_lst[i]
        else:
            for filter in FILTERS:
                param_dict = FILTERS[filter]
                dFC_lst2check = filter_dFC_lst(self.dFC_lst, **param_dict)
                output[filter] = self.assess_similarity(
                    dFC_lst=dFC_lst2check,
                    downsampling_method=downsampling_method,
                    **param_dict,
                )

        return output


#################################################################################################



"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import time

import numpy as np
from joblib import Parallel, delayed

from ..dfc import DFC
from ..time_series import TIME_SERIES
from .base_dfc_method import BaseDFCMethod

################################# Time-Frequency #################################

"""
PyCWT:
Authors
Sebastian Krieger, Nabil Freij, Alexey Brazhe, Christopher Torrence,
Gilbert P. Compo and contributors.

Disclaimer
This module is based on routines provided by C. Torrence and G. P. Compo available
at http://paos.colorado.edu/research/wavelets/, on routines provided by A. Grinsted,
J. Moore and S. Jevrejeva available at
http://noc.ac.uk/using-science/crosswavelet-wavelet-coherence, and on routines
provided by A. Brazhe available at http://cell.biophys.msu.ru/static/swan/.

This software is released under a BSD-style open source license. Please read
the license file for further information. This routine is provided as is
without any express or implied warranties whatsoever.

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.
    dj : float, optional
        Spacing between discrete scales. Default value is 1/12.
        Smaller values will result in better scale resolution, but
        slower calculation and plot.
    s0 : float, optional
        Smallest scale of the wavelet. Default value is 2*dt.
    J : float, optional
        Number of scales less one. Scales range from s0 up to
        s0 * 2**(J * dj), which gives a total of (J + 1) scales.
        Default is J = (log2(N*dt/so))/dj.
    sig : bool
        set to compute significance, default is True
    significance_level (float, optional) :
        Significance level to use. Default is 0.95.
    normalize (boolean, optional) :
        If set to true, normalizes CWT by the standard deviation of
        the signals.

- if n_jobs is None => no parallelization

todo:

- consider COI and edge effect in averaging:
    => should we truncate the time points having at less than 20 freqs as done in Savva et al. ?

"""
import pycwt as wavelet


class TIME_FREQ(BaseDFCMethod):

    def __init__(self, coi_correction=True, **params):

        # assert TF_method in self.TF_methods_name_lst, \
        #     "Time-frequency method not recognized."

        self.logs_ = ""
        self.TPM = []
        self.FCS_ = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = [
            "measure_name",
            "is_state_based",
            "TF_method",
            "coi_correction",
            "n_jobs",
            "verbose",
            "backend",
            "normalization",
            "num_select_nodes",
            "num_time_point",
            "Fs_ratio",
            "noise_ratio",
            "num_realization",
            "session",
        ]
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params["measure_name"] = "Time-Freq"
        self.params["is_state_based"] = False
        self.params["coi_correction"] = coi_correction

    @property
    def measure_name(self):
        return self.params["measure_name"]  # + '_' + self.params['TF_method']

    def coi_correct(self, X, coi, freqs):
        # correct the edge effect in matrix X = [freqs, time] using coi
        # if self.coi_correction=True

        if not self.params["coi_correction"]:
            return X
        periods = 1 / freqs
        periods = np.repeat(periods[:, None], X.shape[1], axis=1)
        coi = np.repeat(coi[None, :], X.shape[0], axis=0)
        X_corrected = np.multiply(X, (coi >= periods))
        return X_corrected

    def WT_dFC(self, Y1, Y2, Fs, J, s0, dj):
        if (
            self.params["TF_method"] == "CWT_mag"
            or self.params["TF_method"] == "CWT_phase_r"
            or self.params["TF_method"] == "CWT_phase_a"
        ):
            # Cross Wavelet Transform
            WT_xy, coi, freqs, _ = wavelet.xwt(
                Y1,
                Y2,
                dt=1 / Fs,
                dj=dj,
                s0=s0,
                J=J,
                significance_level=0.95,
                wavelet="morlet",
                normalize=True,
            )

            if self.params["TF_method"] == "CWT_mag":
                WT_xy_corrected = self.coi_correct(WT_xy, coi, freqs)
                wt = np.abs(np.mean(WT_xy_corrected, axis=0))

            if (
                self.params["TF_method"] == "CWT_phase_r"
                or self.params["TF_method"] == "CWT_phase_a"
            ):
                cosA = np.cos(np.angle(WT_xy))
                sinA = np.sin(np.angle(WT_xy))

                cosA_corrected = self.coi_correct(cosA, coi, freqs)
                sinA_corrected = self.coi_correct(sinA, coi, freqs)

                A = cosA_corrected + sinA_corrected * 1j

                if self.params["TF_method"] == "CWT_phase_r":
                    wt = np.abs(np.mean(A, axis=0))
                else:
                    wt = np.angle(np.mean(A, axis=0))

        if self.params["TF_method"] == "WTC":
            # Wavelet Transform Coherence
            WT_xy, _, coi, freqs, _ = wavelet.wct(
                Y1,
                Y2,
                dt=1 / Fs,
                dj=dj,
                s0=s0,
                J=J,
                sig=False,
                significance_level=0.95,
                wavelet="morlet",
                normalize=True,
            )
            WT_xy_corrected = self.coi_correct(WT_xy, coi, freqs)
            wt = np.abs(np.mean(WT_xy_corrected, axis=0))

        return wt

    def estimate_FCS(self, time_series):

        return self

    def estimate_dFC(self, time_series):
        """
        we assume calc is applied on subjects separately
        """
        assert (
            len(time_series.subj_id_lst) == 1
        ), "this function takes only one subject as input."

        # params
        J = 100  # -1
        s0 = 1  # -1
        dj = 1 / 12  # 1/12

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        WT = np.zeros((time_series.n_time, time_series.n_regions, time_series.n_regions))

        for i in range(time_series.n_regions):
            if self.params["n_jobs"] is None:
                Q = list()
                for j in range(time_series.n_regions):
                    Q.append(
                        self.WT_dFC(
                            Y1=time_series.data[i, :],
                            Y2=time_series.data[j, :],
                            Fs=time_series.Fs,
                            J=J,
                            s0=s0,
                            dj=dj,
                        )
                    )
            else:
                Q = Parallel(
                    n_jobs=self.params["n_jobs"],
                    verbose=self.params["verbose"],
                    backend=self.params["backend"],
                )(
                    delayed(self.WT_dFC)(
                        Y1=time_series.data[i, :],
                        Y2=time_series.data[j, :],
                        Fs=time_series.Fs,
                        J=J,
                        s0=s0,
                        dj=dj,
                    )
                    for j in range(time_series.n_regions)
                )
            WT[:, i, :] = np.array(Q).T

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=WT, TS_info=time_series.info_dict)
        return dFC


################################################################################



"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import time

import numpy as np
from sklearn.cluster import KMeans

from ..dfc import DFC
from ..time_series import TIME_SERIES
from .base_dfc_method import BaseDFCMethod

################################## CAP ##################################

"""
by : web link

Reference: ##

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:
"""


class CAP(BaseDFCMethod):

    def __init__(self, **params):
        self.logs_ = ""
        self.FCS_ = []
        self.mean_act = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = [
            "measure_name",
            "is_state_based",
            "n_states",
            "n_subj_clstrs",
            "normalization",
            "num_subj",
            "num_select_nodes",
            "num_time_point",
            "Fs_ratio",
            "noise_ratio",
            "num_realization",
            "session",
        ]
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params["measure_name"] = "CAP"
        self.params["is_state_based"] = True

    @property
    def measure_name(self):
        return self.params["measure_name"]

    def act_vec2FCS(self, act_vecs):
        FCS_ = list()
        for act_vec in act_vecs:
            FCS_.append(np.multiply(act_vec[:, np.newaxis], act_vec[np.newaxis, :]))
        return np.array(FCS_)

    def cluster_act_vec(self, act_vecs, n_clusters):

        kmeans_ = KMeans(n_clusters=n_clusters, n_init=500).fit(act_vecs)
        kmeans_.cluster_centers_ = kmeans_.cluster_centers_.astype(np.float32)
        act_centroids = kmeans_.cluster_centers_

        return act_centroids, kmeans_

    def estimate_FCS(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        # 2-level clustering
        SUBJECTs = time_series.subj_id_lst
        act_center_1st_level = None
        for subject in SUBJECTs:

            act_vecs = time_series.get_subj_ts(subjs_id=subject).data.T

            # test
            if act_vecs.shape[0] < self.params["n_subj_clstrs"]:
                print(
                    "Number of subject-level clusters cannot be more than time samples! n_subj_clstrs was changed to "
                    + str(act_vecs.shape[0])
                )
                self.params["n_subj_clstrs"] = act_vecs.shape[0]

            act_centroids, _ = self.cluster_act_vec(
                act_vecs=act_vecs, n_clusters=self.params["n_subj_clstrs"]
            )
            if act_center_1st_level is None:
                act_center_1st_level = act_centroids
            else:
                act_center_1st_level = np.concatenate(
                    (act_center_1st_level, act_centroids), axis=0
                )

        group_act_centroids, self.kmeans_ = self.cluster_act_vec(
            act_vecs=act_center_1st_level, n_clusters=self.params["n_states"]
        )
        self.FCS_ = self.act_vec2FCS(group_act_centroids)
        self.Z = self.kmeans_.predict(time_series.data.T.astype(np.float32))

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        assert (
            len(time_series.subj_id_lst) == 1
        ), "this function takes only one subject as input."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        act_vecs = time_series.data.T

        Z = self.kmeans_.predict(act_vecs.astype(np.float32))

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=self.FCS_, FCS_idx=Z, TS_info=time_series.info_dict)
        return dFC


################################################################################



"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import time
from copy import deepcopy

import numpy as np

from ..dfc import DFC
from ..time_series import TIME_SERIES
from .base_dfc_method import BaseDFCMethod
from .sliding_window_clustr import SLIDING_WINDOW_CLUSTR

################################# HMM Discrete #################################

"""

Parameters
    ----------
    Z : numpy.ndarray
        state time course
    M : int
        (num of observations/n_state) of 16
    N : int
        (num of hidden states) of 24
    self.FCC :
        dFC estimated by Clustering which is then used to fit Discrete HMM
    self.FCS_ :
        collection FCS pattern coded in numbers for Discrete HMM

todo:
- two-level hierarchical clustering ?
- find a better name for FCC
"""
# from HMM_discrete import *
from hmmlearn import hmm


class HMM_DISC(BaseDFCMethod):

    def __init__(self, **params):
        self.logs_ = ""
        self.TPM = []
        self.FCS_ = []
        self.mean_act = []
        self.swc = None
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = [
            "measure_name",
            "is_state_based",
            "clstr_base_measure",
            "clstr_distance",
            "sw_method",
            "tapered_window",
            "dhmm_obs_state_ratio",
            "coi_correction",
            "hmm_iter",
            "n_jobs",
            "verbose",
            "backend",
            "n_subj_clstrs",
            "W",
            "window_std",
            "n_overlap",
            "n_states",
            "normalization",
            "num_subj",
            "num_select_nodes",
            "num_time_point",
            "Fs_ratio",
            "noise_ratio",
            "num_realization",
            "session",
        ]
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params["measure_name"] = "DiscreteHMM"
        self.params["is_state_based"] = True

        assert (
            self.params["clstr_base_measure"] in self.base_methods_name_lst
        ), "Base measure not recognized."

    @property
    def measure_name(self):
        return self.params["measure_name"]  # + '_' + self.base_method

    def estimate_FCS(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        # change n_states of swc to n_observations which is dhmm_obs_state_ratio*n_states
        params = deepcopy(self.params)
        params["n_states"] = int(
            self.params["dhmm_obs_state_ratio"] * self.params["n_states"]
        )

        self.swc = SLIDING_WINDOW_CLUSTR(**params)
        self.swc.estimate_FCS(time_series=time_series)

        SUBJECTs = time_series.subj_id_lst
        SWC_dFC = None
        Obs_seq = None
        for subject in SUBJECTs:
            new_dFC = self.swc.estimate_dFC(
                time_series=time_series.get_subj_ts(subjs_id=subject)
            )
            new_dFC_mat = new_dFC.get_dFC_mat()
            new_Obs = new_dFC.FCS_idx_array.reshape(-1, 1)
            if SWC_dFC is None:
                SWC_dFC = new_dFC_mat
                Obs_seq = new_Obs
            else:
                SWC_dFC = np.concatenate((SWC_dFC, new_dFC_mat), axis=0)
                Obs_seq = np.concatenate((Obs_seq, new_Obs), axis=0)

        Models, Scores = [], []
        for i in range(self.params["hmm_iter"]):
            model = hmm.CategoricalHMM(n_components=self.params["n_states"])
            model.fit(Obs_seq)
            score = model.score(Obs_seq)
            Models.append(model)
            Scores.append(score)

        self.hmm_model = Models[np.argmax(Scores)]
        self.Z = self.hmm_model.predict(Obs_seq)
        self.TPM = self.hmm_model.transmat_
        self.EPM = self.hmm_model.emissionprob_

        if len(np.unique(self.Z)) < self.params["n_states"]:
            self.logs_ += "Less number of states were fitted than n_states. \n"

        self.FCS_ = np.zeros(
            (self.params["n_states"], time_series.n_regions, time_series.n_regions)
        )
        for i in range(self.params["n_states"]):
            ids = np.array([int(state == i) for state in self.Z])
            if np.any(ids > 0):
                self.FCS_[i, :, :] = np.average(SWC_dFC, weights=ids, axis=0)

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        assert (
            len(time_series.subj_id_lst) == 1
        ), "this function takes only one subject as input."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        FCC = self.swc.estimate_dFC(time_series=time_series)
        Obs_seq = FCC.FCS_idx_array.reshape(-1, 1)

        Z = self.hmm_model.predict(Obs_seq)

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(
            FCSs=self.FCS_,
            FCS_idx=Z,
            TS_info=time_series.info_dict,
            TR_array=FCC.TR_array,
        )

        return dFC


###################################################################################



"""
Notes:
- Something to be aware of when grouping tests inside classes is that each test has
    a unique instance of the class. Having each test share the same class instance
    would be very detrimental to test isolation and would promote poor test practices.

todo:
- carefully test TS truncate
"""

import time

import hdf5storage
import numpy as np
import scipy.io as sio
from functions.dFC_funcs import *

################################# FUNCTIONS ######################################


def check_symmetric(A, rtol=1e-05, atol=1e-08):

    for t in range(A.shape[0]):
        if not np.allclose(A[t, :, :], A[t, :, :].T, rtol=rtol, atol=atol):
            return False
    return True


def sign_maintenance(C, C_z):
    sign_check = True
    for t in range(C.shape[0]):
        slice_C = C[t, :, :]
        slice_C_z = C_z[t, :, :]
        slice_C = slice_C[np.where(~np.eye(slice_C.shape[0], dtype=bool))]
        slice_C_z = slice_C_z[np.where(~np.eye(slice_C_z.shape[0], dtype=bool))]
        if not np.all(
            np.sign(slice_C[slice_C_z != 0]) == np.sign(slice_C_z[slice_C_z != 0])
        ):
            sign_check = False
    return sign_check


################################# UNIT TESTs ######################################
class Test_dFC_mat_normalize:

    def test_constant(self):
        C = 2 * np.ones((120, 333, 333))
        C_z = dFC_mat_normalize(C, global_normalization=False, threshold=0.0)
        assert np.all(C_z == 0), "constant matrix normalization failed."
        C_z = dFC_mat_normalize(C, global_normalization=True, threshold=0.0)
        assert np.all(C_z == 0), "constant matrix normalization failed."

    def test_random_C(self):
        C = -1 + 2 * np.random.randn(120, 333, 333)
        for i in range(C.shape[1]):
            C[:, i, i] = 1
        C_z = dFC_mat_normalize(C, global_normalization=False, threshold=0.0)

        assert sign_maintenance(
            C=C, C_z=C_z
        ), "random matrix normalization sign maintenance failed."

        assert np.all(C_z <= 1) and np.all(
            C_z >= -1
        ), "random matrix normalization failed."

    def test_cov_C(self):
        TS = np.random.randn(333, 200)
        C = list()
        for l in range(100):
            C.append(np.cov(TS[:, l : l + 100], bias=False))
        C = np.array(C)
        assert check_symmetric(C), "Covariance matrix Not symmetric."

        C_z = dFC_mat_normalize(C, global_normalization=False, threshold=0.0)

        assert np.all(C_z <= 1) and np.all(
            C_z >= -1
        ), "Covariance matrix normalization failed."
        assert sign_maintenance(
            C=C, C_z=C_z
        ), "Covariance matrix normalization sign maintenance failed."
        assert check_symmetric(
            C_z
        ), "Covariance matrix normalization failed. Not symmetric."

    # def test_threshold(self):
    #     C = -1 + 2 * np.random.randn(120, 333, 333)
    #     for i in range(C.shape[1]):
    #         C[:,i,i] = 1
    #     C_z = dFC_mat_normalize(C, global_normalization=False, threshold=0.0)
    #     assert time_series.n_regions==333, \
    #         "n_regions property failed."


class Test_TIME_SERIES_CLASS:

    data = np.random.randn(333, 1200)
    time_array = np.random.randn(1200)
    time_array = np.sort(time_array)
    data2 = np.random.randn(333, 1200)
    time_array2 = np.random.randn(1200)
    time_array2 = np.sort(time_array2)

    def test_n_regions(self):
        time_series = TIME_SERIES(data=self.data, subj_id="1", Fs=2)
        assert time_series.n_regions == 333, "n_regions property failed."

    def test_get_subj_ts(self):

        ########## normal append_ts ##############
        time_series = TIME_SERIES(
            data=self.data, subj_id="1", Fs=2, time_array=self.time_array2
        )
        time_series.append_ts(
            new_time_series=self.data2, subj_id="2", time_array=self.time_array2
        )

        assert time_series.n_time_ == 2400, "n_time_ property failed."
        assert time_series.n_time == 2400, "n_time property failed."
        assert time_series.n_regions_ == 333, "n_regions_ property failed."
        assert time_series.n_regions == 333, "n_regions property failed."

        assert time_series.interval_ == list(
            range(time_series.n_time_)
        ), "interval property failed."
        assert time_series.interval.tolist()[0] == list(
            range(time_series.n_time_)
        ), "interval property failed."

        assert np.all(
            time_series.get_subj_ts("1").data_ == self.data
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("1").data == self.data
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").data_ == self.data2
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").data == self.data2
        ), "get_subj_ts method failed."

        assert (
            np.unique(time_series.get_subj_ts("1").subj_id_array) == "1"
        ), "get_subj_ts method failed."
        assert (
            np.unique(time_series.get_subj_ts("2").subj_id_array) == "2"
        ), "get_subj_ts method failed."

        ########## append_ts after node selection ##############
        time_series = TIME_SERIES(
            data=self.data, subj_id="1", Fs=2, time_array=self.time_array2
        )
        nodes_idx = np.random.choice(range(time_series.n_regions), size=50, replace=False)
        nodes_idx.sort()
        time_series.select_nodes(nodes_idx=nodes_idx)
        time_series.append_ts(
            new_time_series=self.data2, subj_id="2", time_array=self.time_array2
        )

        assert time_series.n_regions_ == 333, "n_regions_ property failed."
        assert time_series.n_regions == 50, "n_regions property failed."

        assert np.all(
            time_series.get_subj_ts("1").data_ == self.data[nodes_idx, :]
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("1").data == self.data[nodes_idx, :]
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("1").get_subj_ts("1").get_subj_ts("1").data
            == self.data[nodes_idx, :]
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").data_ == self.data2[nodes_idx, :]
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").data == self.data2[nodes_idx, :]
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").get_subj_ts("2").get_subj_ts("2").data
            == self.data2[nodes_idx, :]
        ), "get_subj_ts method failed."

        ########## append_ts after truncate ##############
        time_series = TIME_SERIES(
            data=self.data, subj_id="1", Fs=2, time_array=self.time_array2
        )
        time_series.truncate(start_point=10, end_point=100)
        time_series.append_ts(
            new_time_series=self.data2, subj_id="2", time_array=self.time_array2
        )

        assert time_series.n_regions == 333, "n_regions_ property failed."
        assert time_series.n_time_ == 2400, "n_time_ property failed."
        assert time_series.n_time == 2400, "n_time property failed."

        assert np.all(
            time_series.get_subj_ts("1").data_ == self.data
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("1").data == self.data
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("1").get_subj_ts("1").get_subj_ts("1").data
            == self.data
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").data_ == self.data2
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").data == self.data2
        ), "get_subj_ts method failed."
        assert np.all(
            time_series.get_subj_ts("2").get_subj_ts("2").get_subj_ts("2").data
            == self.data2
        ), "get_subj_ts method failed."

    def test_n_time(self):
        time_series = TIME_SERIES(data=self.data, subj_id="1", Fs=2)
        assert time_series.n_time_ == 1200, "n_time property failed."
        assert time_series.n_time == 1200, "n_time property failed."
        time_series.truncate(start_point=10, end_point=100)
        assert time_series.n_time_ == 1200, "TRUNCATE: n_time property failed."
        assert time_series.n_time == 91, "TRUNCATE: n_time property failed."
        time_series.truncate()
        assert time_series.n_time_ == 1200, "TRUNCATE_RESET: n_time property failed."
        assert time_series.n_time == 1200, "TRUNCATE_RESET: n_time property failed."

    def test_append_ts(self):
        time_series = TIME_SERIES(data=self.data, subj_id="1", Fs=2)
        time_series.append_ts(new_time_series=np.random.randn(333, 1200), subj_id="2")
        assert time_series.n_time == 2400, "append_ts failed."
        assert time_series.n_regions == 333, "append_ts failed."

    def test_node_lst_after_append(self):
        time_series = TIME_SERIES(data=self.data, subj_id="1", Fs=2)
        nodes_idx = np.random.choice(333, size=50, replace=False)
        # nodes_idx.sort()
        time_series.select_nodes(nodes_idx=nodes_idx)
        time_series.append_ts(new_time_series=np.random.randn(333, 1200), subj_id="2")

        assert (
            time_series.n_regions == 50
        ), "keeping node selection after append_ts failed."

        assert np.all(
            time_series.nodes_lst == np.array(nodes_idx)[:, np.newaxis]
        ), "keeping node selection after append_ts failed."

        time_series.select_nodes(nodes_idx=None)
        assert (
            time_series.n_regions == 333
        ), "resetting node selection after select_nodes(None) failed."


class Test_DFCM_CLASS:

    FCPs = np.random.rand(12, 333, 333)
    FCP_idx = np.random.randint(low=0, high=12, size=(27,))
    TR_array = np.arange(43, 1200, 44)

    FCPs2 = np.random.rand(12, 333, 333)
    FCP_idx2 = np.random.randint(low=0, high=12, size=(27,))
    TR_array2 = np.arange(43, 1200, 44)

    def test_get_dFC_mat(self):
        dFCM = DFCM()
        dFCM.add_FCP(
            FCPs=self.FCPs,
            FCP_idx=self.FCP_idx,
            subj_id_array=[1] * len(self.FCP_idx),
            TR_array=self.TR_array,
        )

        assert np.all(
            dFCM.get_dFC_mat(TRs=[131]) == self.FCPs[self.FCP_idx[2]]
        ), "get_dFC_mat method failed."

        idx = np.random.randint(low=0, high=27, size=(10,))
        assert np.all(
            dFCM.get_dFC_mat(TRs=self.TR_array[idx]) == self.FCPs[self.FCP_idx[idx]]
        ), "get_dFC_mat method failed."

    def test_add_FCP_multiple(self):
        ########## add_FCP multiple ##############
        dFCM = DFCM()
        dFCM.add_FCP(
            FCPs=self.FCPs,
            FCP_idx=self.FCP_idx,
            subj_id_array=[1] * len(self.FCP_idx),
            TR_array=self.TR_array,
        )
        assert dFCM.TR_array[2] == 131, "add_FCP method failed."
        assert dFCM.n_time == 27, "add_FCP method failed."

    def test_add_FCP_single(self):
        ########## add_FCP one-by-one ##############
        # for SW class

        dFCM = DFCM()
        for i in range(12):
            dFCM.add_FCP(
                FCPs=self.FCPs[i], subj_id_array="1", TR_array=self.TR_array[i : i + 1]
            )

        assert dFCM.TR_array[2] == 131, "add_FCP method failed."
        assert dFCM.n_time == 12, "add_FCP method failed."

        idx = np.random.randint(low=0, high=12, size=(10,))
        assert np.all(
            dFCM.get_dFC_mat(TRs=self.TR_array[idx]) == self.FCPs[idx]
        ), "get_dFC_mat method failed."

    def test_two(self):
        dFCM = DFCM()
        dFCM.add_FCP(
            FCPs=self.FCPs,
            FCP_idx=self.FCP_idx,
            subj_id_array=[1] * len(self.FCP_idx),
            TR_array=self.TR_array,
        )
        assert dFCM.n_time == 27


# class Test_methods:

#     Fs = 2
#     data = np.random.randn(50, 300)
#     data2 = np.random.randn(50, 300)
#     time_BOLD = 1/Fs + np.arange(0, data.shape[1]/Fs, 1/Fs)

#     n_states = 2
#     n_subj_clstrs = 4
#     n_hid_states = 3
#     n_overlap = 0.5
#     W_sw = 44 # in seconds, 44, choose even Ws!?
#     n_jobs = None
#     verbose=0

#     params = {'W': int(W_sw*Fs), 'n_overlap': n_overlap, \
#     'n_states': n_states, 'n_subj_clstrs': n_subj_clstrs, 'n_hid_states': n_hid_states, \
#     'n_jobs': n_jobs, 'verbose': verbose, 'backend': 'loky' \
#             }

#     hmm_cont = HMM_CONT(**params)
#     windowless = WINDOWLESS(**params)

#     sw_pc = SLIDING_WINDOW(sw_method='pear_corr', **params)
#     sw_mi = SLIDING_WINDOW(sw_method='MI', **params)
#     sw_gLasso = SLIDING_WINDOW(sw_method='GraphLasso', **params)

#     time_freq_cwt = TIME_FREQ(method='CWT_mag', **params)
#     time_freq_cwt_r = TIME_FREQ(method='CWT_phase_r', **params)
#     time_freq_wtc = TIME_FREQ(method='WTC', **params)

#     swc_pc = SLIDING_WINDOW_CLUSTR(sw_method='pear_corr', **params)
#     swc_mi = SLIDING_WINDOW_CLUSTR(sw_method='MI', **params)
#     # swc_gLasso = SLIDING_WINDOW_CLUSTR(sw_method='GraphLasso', **params)

#     hmm_disc_pc = HMM_DISC(sw_method='pear_corr', **params)
#     hmm_disc_mi = HMM_DISC(sw_method='MI', **params)
#     # hmm_disc_gLasso = HMM_DISC(sw_method='GraphLasso', **params)

#     time_series = TIME_SERIES(data=data, Fs=Fs, subj_id='1', time_array=time_BOLD, TS_name='BOLD Simulation')

#     def test_one(self):

#         MEASURES = [
#             self.hmm_cont, \
#             self.windowless, \
#             self.sw_pc, \
#             self.sw_mi, \
#             self.sw_gLasso, \
#             self.time_freq_cwt, \
#             self.time_freq_cwt_r, \
#             self.time_freq_wtc, \
#             self.swc_pc, \
#             # self.swc_mi, \
#             # self.swc_gLasso, \
#             self.hmm_disc_pc,\
#             # self.hmm_disc_mi \
#             # self.hmm_disc_gLasso, \
#                     ]

#         for measure in MEASURES:
#             if measure.is_state_based:
#                 measure.estimate_FCS(time_series=self.time_series)
#             dFCM = measure.estimate_dFCM(time_series=self.time_series)

#             assert not dFCM is None, \
#                 measure.measure_name + " has None dFCM."

#             assert dFCM.n_regions==self.time_series.n_regions, \
#                 measure.measure_name + " has n_regions mismatch."
################################# INTEGRATION TESTs ######################################


################################# MAIN ######################################

# if __name__ == '__main__':
#     TestNumber.test_two(x=2)



"""
Implementation of functions for loading fmri data.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import os
from copy import deepcopy

import h5py
import numpy as np

from .dfc_utils import intersection, label2network
from .time_series import TIME_SERIES

################################# DATA_LOADER functions ######################################


def find_subj_list(data_root):
    """
    find the list of subjects in data_root
    the files must follow the format: sub-subjectID
    only these files should be in the data_root
    """
    if data_root[-1] != "/":
        data_root += "/"

    ALL_FILES = os.listdir(data_root)
    FOLDERS = [item for item in ALL_FILES if os.path.isdir(data_root + item)]

    FOLDERS.sort()
    SUBJECTS = list()
    for s in FOLDERS:
        if "sub-" in s:
            SUBJECTS.append(s)
    SUBJECTS.sort()

    print(f"{len(SUBJECTS)} subjects were found. ")

    return SUBJECTS


def load_from_array(subj_id2load=None, **params):
    """
    load fMRI data from numpy or mat files
    input time_series.shape must be (time, roi)
    returns a dictionary of TIME_SERIES objects
    each corresponding to a session

    - if the file_name is a .mat file, it will be loaded using h5py
      if the file_name is a .npy file, it will be loaded using np.load

    - the roi locations should be in the same folder and a .npy file
      with the name: params['roi_locs_file']
      it must

    - and the roi labels should be in the same folder and a .npy file
      with the name: params['roi_labels_file']
      it must be a list of strings

    - labels should be in the format: Hemisphere_Network_ID
        ow, the network2include will not work properly
    """

    SESSIONs = params["SESSIONs"]  # list of sessions
    if subj_id2load is None:
        SUBJECTS = find_subj_list(params["data_root"])
    else:
        SUBJECTS = [subj_id2load]

    # LOAD Region Location DATA
    locs = np.load(
        params["data_root"] + params["roi_locs_file"], allow_pickle="True"
    ).item()
    locs = locs["locs"]

    # LOAD Region Labels DATA
    labels = np.load(
        params["data_root"] + params["roi_labels_file"], allow_pickle="True"
    ).item()
    labels = labels["labels"]

    assert type(locs) is np.ndarray, "locs must be a numpy array"
    assert type(labels) is list, "labels must be a list"
    assert locs.shape[0] == len(labels), "locs and labels must have the same length"
    assert locs.shape[1] == 3, "locs must have 3 columns"

    # apply networks2include
    # if params['networks2include'] is None, all the regions will be included
    if not params["networks2include"] is None:
        nodes2include = [
            i
            for i, x in enumerate(labels)
            if label2network(x) in params["networks2include"]
        ]
    else:
        nodes2include = [i for i, x in enumerate(labels)]
    locs = locs[nodes2include, :]
    labels = [x for node, x in enumerate(labels) if node in nodes2include]

    BOLD = {}
    for session in SESSIONs:
        BOLD[session] = None
        for subject in SUBJECTS:

            subj_fldr = subject + "_" + session

            # LOAD BOLD Data

            if params["file_name"][params["file_name"].find(".") :] == ".mat":
                with h5py.File(
                    params["data_root"] + subj_fldr + "/" + params["file_name"], "r"
                ) as f:
                    DATA = {k: np.array(f[k]) for k in f.keys()}
            elif params["file_name"][params["file_name"].find(".") :] == ".npy":
                DATA = np.load(
                    params["data_root"] + subj_fldr + "/" + params["file_name"],
                    allow_pickle="True",
                ).item()
            time_series = DATA["ROI_data"]  # time_series.shape = (time, roi)

            # change time_series.shape to (roi, time)
            time_series = time_series.T

            # apply networks2include
            time_series = time_series[nodes2include, :]

            if BOLD[session] is None:
                BOLD[session] = TIME_SERIES(
                    data=time_series,
                    subj_id=subject,
                    Fs=params["Fs"],
                    locs=locs,
                    node_labels=labels,
                    TS_name="BOLD Real",
                    session_name=session,
                )
            else:
                BOLD[session].append_ts(new_time_series=time_series, subj_id=subject)

        print("*** Session " + session + ": ")
        print(
            "number of regions= "
            + str(BOLD[session].n_regions)
            + ", number of time points= "
            + str(BOLD[session].n_time)
        )

    return BOLD


def nifti2array(nifti_file, confound_strategy="none", standardize=False, n_rois=100):
    """
    this function uses nilearn maskers to extract
    BOLD signals from nifti files
    For now it only works with schaefer atlas,
    but you can set the number of rois to extract
    {100, 200, 300, 400, 500, 600, 700, 800, 900, 1000}

    returns a numpy array of shape (time, roi)
    and labels and locs of rois

    confound_strategy:
        'none': no confounds are used
        'no_motion': motion parameters are used
        'no_motion_no_gsr': motion parameters are used
                            and global signal regression
                            is applied.
    """
    from nilearn import datasets
    from nilearn.interfaces.fmriprep import load_confounds
    from nilearn.maskers import NiftiLabelsMasker
    from nilearn.plotting import find_parcellation_cut_coords

    parc = datasets.fetch_atlas_schaefer_2018(n_rois=n_rois)
    atlas_filename = parc.maps
    labels = parc.labels
    # The list of labels does not contain ‘Background’ by default.
    # To have proper indexing, you should either manually add ‘Background’ to the list of labels:
    # Prepend background label
    labels = np.insert(labels, 0, "Background")

    # extract locs
    # test!
    # check if order is the same as labels
    locs, labels_ = find_parcellation_cut_coords(
        atlas_filename, background_label=0, return_label_names=True
    )

    # create the masker for extracting time series
    masker = NiftiLabelsMasker(
        labels_img=atlas_filename,
        labels=labels,
        resampling_target="data",
        standardize=standardize,
    )

    labels = np.delete(labels, 0)  # remove the background label
    labels = [label.decode() for label in labels]

    ### extract the timeseries
    if confound_strategy == "none":
        time_series = masker.fit_transform(nifti_file)
    elif confound_strategy == "no_motion":
        confounds_simple, sample_mask = load_confounds(
            nifti_file,
            strategy=["high_pass", "motion", "wm_csf"],
            motion="basic",
            wm_csf="basic",
        )
        time_series = masker.fit_transform(
            nifti_file, confounds=confounds_simple, sample_mask=sample_mask
        )
    elif confound_strategy == "no_motion_no_gsr":
        confounds_simple, sample_mask = load_confounds(
            nifti_file,
            strategy=["high_pass", "motion", "wm_csf", "global_signal"],
            motion="basic",
            wm_csf="basic",
            global_signal="basic",
        )
        time_series = masker.fit_transform(
            nifti_file, confounds=confounds_simple, sample_mask=sample_mask
        )

    return time_series, labels, locs


def nifti2timeseries(
    nifti_file,
    n_rois,
    Fs,
    subj_id,
    confound_strategy="none",
    standardize=False,
    TS_name=None,
    session=None,
):
    """
    this function is only for single subject and single session data loading
    it uses nilearn maskers to extract ROI signals from nifti files
    and returns a TIME_SERIES object

    For now it only works with schaefer atlas,
    but you can set the number of rois to extract
    {100, 200, 300, 400, 500, 600, 700, 800, 900, 1000}
    """
    time_series, labels, locs = nifti2array(
        nifti_file=nifti_file,
        confound_strategy=confound_strategy,
        standardize=standardize,
        n_rois=n_rois,
    )

    assert type(locs) is np.ndarray, "locs must be a numpy array"
    assert type(labels) is list, "labels must be a list"
    assert locs.shape[0] == len(labels), "locs and labels must have the same length"
    assert locs.shape[1] == 3, "locs must have 3 columns"

    # change time_series.shape to (roi, time)
    time_series = time_series.T

    if TS_name is None:
        TS_name = subj_id + " time series"

    BOLD = TIME_SERIES(
        data=time_series,
        subj_id=subj_id,
        Fs=Fs,
        locs=locs,
        node_labels=labels,
        TS_name=TS_name,
        session_name=session,
    )

    return BOLD


def multi_nifti2timeseries(
    nifti_files_list,
    subj_id_list,
    n_rois,
    Fs,
    confound_strategy="none",
    standardize=False,
    TS_name=None,
    session=None,
):
    """
    loading data of multiple subjects, but single session, from their nifti files
    """
    BOLD_multi = None
    for subj_id, nifti_file in zip(subj_id_list, nifti_files_list):
        if BOLD_multi is None:
            BOLD_multi = nifti2timeseries(
                nifti_file=nifti_file,
                n_rois=n_rois,
                Fs=Fs,
                subj_id=subj_id,
                confound_strategy=confound_strategy,
                standardize=standardize,
                TS_name=TS_name,
                session=session,
            )
        else:
            BOLD_multi.concat_ts(
                nifti2timeseries(
                    nifti_file=nifti_file,
                    n_rois=n_rois,
                    Fs=Fs,
                    subj_id=subj_id,
                    confound_strategy=confound_strategy,
                    standardize=standardize,
                    TS_name=TS_name,
                    session=session,
                )
            )
    return BOLD_multi


def load_TS(
    data_root,
    file_name,
    SESSIONs,
    subj_id2load=None,
    task=None,
    run=None,
):
    """
    load a TIME_SERIES object from a .npy file
    if SESSIONs is a list, it will load all the sessions,
        if it is a string, it will load that session
    if subj_id2load is None, it will load all the subjects
    file_name: name of the file to load
        format example: {subj_id}_{task}_{run}_time-series.npy
        (keep the {} for the variables)
    """
    # check if SESSIONs is a list or a string
    flag = False
    if type(SESSIONs) is str:
        SESSIONs = [SESSIONs]
        flag = True

    if subj_id2load is None:
        SUBJECTS = find_subj_list(data_root)
    else:
        assert "sub-" in subj_id2load, "subj_id2load must start with 'sub-'"
        SUBJECTS = [subj_id2load]

    TS = {}
    for session in SESSIONs:
        TS[session] = None
        for subj in SUBJECTS:
            subj_fldr = subj
            # make the file_name
            TS_file = deepcopy(file_name)
            if "{subj_id}" in file_name:
                TS_file = TS_file.replace("{subj_id}", subj)
            if "{task}" in file_name:
                assert task is not None, "task must be provided"
                TS_file = TS_file.replace("{task}", task)
            if "{run}" in file_name:
                assert run is not None, "run must be provided"
                TS_file = TS_file.replace("{run}", run)

            try:
                time_series = np.load(
                    f"{data_root}/{subj_fldr}/{TS_file}", allow_pickle="True"
                ).item()
            except FileNotFoundError:
                print(f"File {TS_file} not found for {subj}")
                continue

            if TS[session] is None:
                TS[session] = time_series
            else:
                TS[session].concat_ts(time_series)

    if flag:
        return TS[SESSIONs[0]]
    return TS


####################################################################################################################################



import os
import time

import hdf5storage
import numpy as np
import scipy.io as sio
from functions.dFC_funcs import *

os.environ["MKL_NUM_THREADS"] = "64"
os.environ["NUMEXPR_NUM_THREADS"] = "64"
os.environ["OMP_NUM_THREADS"] = "64"

print(
    "################################# CODE started running ... #################################"
)

################################# Parameters #################################

###### DATA PARAMETERS ######

output_root = "./"

# DATA_type is either 'sample' or 'Gordon' or 'simulated' or 'ICA'
params_data_load = {
    "DATA_type": "Gordon",
    "SESSIONs": ["Rest1_LR", "Rest1_RL", "Rest2_LR", "Rest2_RL"],
    "networks2include": [
        "Auditory",
        "CinguloOperc",
        "Default",
        "DorsalAttn",
        "FrontoParietal",
        "MedialParietal",
        "ParietoOccip",
        "SMhand",
        "SMmouth",
        "Salience",
        "VentralAttn",
        "Visual",
    ],
    "data_root_simul": "./../../../../DATA/TVB data/",
    "data_root_sample": "./sampleDATA/",
    "data_root_gordon": "./../../../../DATA/HCP/HCP_Gordon/",
    "data_root_ica": "./../../../../DATA/HCP/HCP_PTN1200/node_timeseries/3T_HCP1200_MSMAll_d50_ts2/",
}

###### MEASUREMENT PARAMETERS ######

# W is in sec

params_methods = {
    # Sliding Parameters
    "W": 44,
    "n_overlap": 0.5,
    "sw_method": "pear_corr",
    "tapered_window": True,
    # TIME_FREQ
    "TF_method": "WTC",
    # CLUSTERING AND DHMM
    "clstr_base_measure": "SlidingWindow",
    # HMM
    "hmm_iter": 30,
    "dhmm_obs_state_ratio": 16 / 24,
    # State Parameters
    "n_states": 12,
    "n_subj_clstrs": 20,
    # Parallelization Parameters
    "n_jobs": 2,
    "verbose": 0,
    "backend": "loky",
    # SESSION
    "session": "Rest1_LR",
    # Hyper Parameters
    "normalization": True,
    "num_subj": 395,
    "num_select_nodes": 96,
    "num_time_point": 1200,
    "Fs_ratio": 1.00,
    "noise_ratio": 0.00,
    "num_realization": 1,
}

###### HYPER PARAMETERS ALTERNATIVE ######

MEASURES_name_lst = [
    "SlidingWindow",
    "Time-Freq",
    "CAP",
    "ContinuousHMM",
    "Windowless",
    "Clustering",
    "DiscreteHMM",
]

alter_hparams = {
    "session": ["Rest1_RL", "Rest2_LR", "Rest2_RL"],
    # 'n_overlap': [0, 0.25, 0.75, 1],
    # 'n_states': [6, 16],
    # # 'normalization': [],
    # 'num_subj': [50, 100, 200],
    # 'num_select_nodes': [30, 50, 333],
    # 'num_time_point': [800, 1000],
    # 'Fs_ratio': [0.50, 0.75, 1.5],
    # 'noise_ratio': [1.00, 2.00, 3.00],
    # 'num_realization': []
}

###### dFC ANALYZER PARAMETERS ######

params_dFC_analyzer = {
    # Parallelization Parameters
    "n_jobs": None,
    "verbose": 0,
    "backend": "loky",
}


################################# LOAD DATA #################################

data_loader = DATA_LOADER(**params_data_load)
BOLD = data_loader.load()

################################# Visualize BOLD #################################

# for session in BOLD:
#     BOLD[session].visualize(start_time=0, end_time=50, nodes_lst=list(range(10)), \
#         save_image=params_dFC_analyzer['save_image'], output_root=output_root+'BOLD_signal_'+session)

################################# Measures of dFC #################################

dFC_analyzer = DFC_ANALYZER(
    analysis_name="reproducibility assessment", **params_dFC_analyzer
)

MEASURES_lst = dFC_analyzer.measures_initializer(
    MEASURES_name_lst, params_methods, alter_hparams
)

tic = time.time()
print("Measurement Started ...")

################################# estimate FCS #################################

task_id = int(os.getenv("SGE_TASK_ID"))
MEASURE_id = task_id - 1  # SGE_TASK_ID starts from 1 not 0


if MEASURE_id >= len(MEASURES_lst):
    print("MEASURE_id out of MEASURES_lst ")
else:
    measure = MEASURES_lst[MEASURE_id]

    print("FCS estimation started...")

    time_series = BOLD[measure.params["session"]]
    if measure.is_state_based:
        measure.estimate_FCS(time_series=time_series)

    # dFC_analyzer.estimate_group_FCS(time_series_dict=BOLD)
    print("FCS estimation done.")

    print(f"Measurement required {time.time() - tic:0.3f} seconds.")

    # Save
    np.save(output_root + "fitted_MEASURES/MEASURE_" + str(MEASURE_id) + ".npy", measure)
    np.save(output_root + "dFC_analyzer.npy", dFC_analyzer)
    np.save(output_root + "data_loader.npy", data_loader)

#################################################################################



"""
dFC class

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np

from .dfc_utils import (
    SW_downsample,
    node_info2network,
    node_labels2networks,
    rank_norm_dFC_dict,
    visualize_conn_mat_dict,
)

################################# DFC class ######################################

"""
Parameters
    ----------
    TR_array : an array labeling
        timepoints by their TRs
        starts from 0

Variables
    ----------
    FCSs : Functional Connecitivity
        States patterns
    FCS_idx : the  index of the
        FCS that corresponds to each
        timepoint


todo:
-
"""


class DFC:
    def __init__(self, measure=None):

        # assert not measure is None, \
        #     "measure arg must be provided."
        self.measure_ = measure
        self.FCSs_ = None  # is a dict
        self.FCS_idx_ = None  # is a dict
        # info of the time series used for dFC estimation
        self.TS_info_ = None
        self.TR_array_ = None
        self.n_regions_ = None
        self.n_time_ = -1

    @classmethod
    def from_numpy(cls, array=None):
        pass

    @property
    def measure(self):
        return self.measure_

    @property
    def TR_array(self):
        return self.TR_array_.astype(int)

    @property
    def TR_keys(self):
        TRs_lst = list()
        for TR in self.TR_array:
            TRs_lst.append("TR" + str(TR))
        return TRs_lst

    @property
    def n_regions(self):
        return self.n_regions_

    @property
    def n_time(self):
        return self.n_time_

    # test this
    @property
    def FCSs(self):
        return self.FCSs_

    # test this
    @property
    def FCS_idx(self):
        return self.FCS_idx_

    # test this
    @property
    def FCS_idx_array(self):
        return np.array(
            [
                int(self.FCS_idx[TR][self.FCS_idx[TR].find("S") + 1 :]) - 1
                for TR in self.FCS_idx
            ]
        )

    @property
    def TS_info(self):
        # info of the time series used for dFC estimation
        return self.TS_info_

    # test
    def state_TC(self, TRs=None, state_match=False, state_match_dict=None):
        # returns a np array of state indices over TRs in TRs

        if TRs is None:
            TRs = self.TR_array

        if not type(TRs[0]) is str:
            TRs_lst = list()
            for TR in TRs:
                TRs_lst.append("TR" + str(TR))
        else:
            TRs_lst = TRs

        state_TC = list()
        for key in self.FCS_idx:
            if key in TRs_lst:
                state = self.FCS_idx[key]
                if state_match:
                    match = state_match_dict["FCS_match"][state]["match"]
                    state_TC.append(int(match[match.find("FCS") + 3 :]))
                else:
                    state_TC.append(int(state[state.find("FCS") + 3 :]))

        state_TC = np.array(state_TC)
        return state_TC

    # test
    def state_act_dict(self, TRs=None):
        # returns a dict including each FCS and its activation times
        # the TRs arg can be used to set a common set of TRs

        if TRs is None:
            TRs = self.TR_array

        TRs_lst = list()
        for TR in TRs:
            TRs_lst.append("TR" + str(TR))

        state_act_dict = {}
        state_act_dict["state_TC"] = {}
        state_act_dict["TR_array"] = TRs
        for FCS_key in self.FCSs:
            state_act_dict["state_TC"][FCS_key] = {}
            state_act_dict["state_TC"][FCS_key]["FCS"] = self.FCSs[FCS_key]
            state_act_dict["state_TC"][FCS_key]["act_TC"] = np.zeros((len(TRs),))
        t = 0
        for TR in self.FCS_idx:
            if TR in TRs_lst:
                state_act_dict["state_TC"][self.FCS_idx[TR]]["act_TC"][t] = 1
                t = t + 1
        assert t == len(TRs), "error!"

        return state_act_dict

    # test
    def dFC2dict(self, TRs=None):
        # return dFC samples as a dictionary
        if TRs is None:
            TRs = self.TR_array
        if type(TRs) is list:
            TRs = np.array(TRs)
        TRs = TRs.astype(int)
        dFC_mat = self.get_dFC_mat(TRs=TRs)
        dFC_dict = {}
        for k, TR in enumerate(TRs):
            dFC_dict[f"TR{TR}"] = dFC_mat[k, :, :]
        return dFC_dict

    # test this
    def get_dFC_mat(self, TRs=None, num_samples=None):
        """
        get dFC matrices corresponding to
        the specified TRs
        TRs should be list/ndarray not necessarily in order ?
        if num_samples specified, it will downsample
        TRs to reach that number of samples and will also
        return picked TRs
        if num_samples > len(TRs) -> picks all TRs
        """

        if TRs is None:
            TRs = self.TR_array

        if type(TRs) is np.int32 or type(TRs) is np.int64 or type(TRs) is int:
            TRs = [TRs]

        if not num_samples is None:
            if num_samples < len(TRs):
                TRs = TRs[
                    np.linspace(0, len(TRs), num_samples, endpoint=False, dtype=int)
                ]

        dFC_mat = list()
        for TR in TRs:
            dFC_mat.append(self.FCSs[self.FCS_idx[f"TR{TR}"]])

        dFC_mat = np.array(dFC_mat)

        if num_samples is None:
            return dFC_mat
        else:
            return dFC_mat, TRs

    def SWed_dFC_mat(self, W=None, n_overlap=None, tapered_window=False):
        """
        the time samples will be picked after
        averaging over a window which slides
        W is in sec
        """
        dFC_mat = self.get_dFC_mat()

        # method not applicable to SW-based methods
        if "sw_method" in self.measure.info:
            return dFC_mat

        dFC_mat_new = SW_downsample(
            data=dFC_mat,
            Fs=self.TS_info["Fs"],
            W=W,
            n_overlap=n_overlap,
            tapered_window=tapered_window,
        )

        return dFC_mat_new

    def set_dFC(self, FCSs, FCS_idx=None, TS_info=None, TR_array=None):
        """
        FCSs: a 3D numpy array of FC matrices with shape (n_time, n_regions, n_regions)
        FCS_idx: a list of indices that correspond to each FC matrix in FCSs over time
        """

        if len(FCSs.shape) == 2:
            FCSs = np.expand_dims(FCSs, axis=0)

        if FCS_idx is None:
            # usually for state-free methods like sliding window when we don't have FCSs
            # we consider each FC a FCS
            FCS_idx = np.arange(start=0, stop=FCSs.shape[0], step=1, dtype=int)

        if type(FCS_idx) is list:
            FCS_idx = np.array(FCS_idx)

        if len(FCS_idx.shape) > 1:
            FCS_idx = np.squeeze(FCS_idx)

        assert FCSs.shape[1] == FCSs.shape[2], "FC matrices must be square."

        assert (
            self.n_time == -1
        ), "why n_time is not -1 ? Are you adding a dFC to an existing dFC ?"

        if TR_array is None:
            # self.n_time is -1 at first. if it is not -1, it means that a dFC is already set and
            # we are adding a new dFC to it.
            TR_array = np.arange(
                start=self.n_time + 1,
                stop=self.n_time + len(FCS_idx) + 1,
                step=1,
                dtype=int,
            )

        assert np.sum(np.abs(np.sort(TR_array) - TR_array)) == 0.0, "TRs not sorted !"

        # the input FCS_idx is ranged from 0 to len(FCS)-1 but we shift it to 1 to len(FCS)
        self.FCSs_ = {}
        for i, FCS in enumerate(FCSs):
            self.FCSs_[f"FCS{i + 1}"] = FCS

        self.FCS_idx_ = {}
        for i, idx in enumerate(FCS_idx):
            self.FCS_idx_[f"TR{TR_array[i]}"] = f"FCS{idx + 1}"  # "FCS" + str(idx + 1)

        self.TS_info_ = TS_info
        self.n_regions_ = FCSs.shape[1]
        self.n_time_ = len(self.FCS_idx_)
        self.TR_array_ = TR_array

    def visualize_dFC(
        self,
        TRs=None,
        normalize=False,
        show_networks=False,
        rank_norm=False,
        threshold=0.0,
        fix_lim=False,
        save_image=False,
        output_root=None,
    ):

        assert not self.measure is None, "Measure is not provided."

        if TRs is None:
            TRs = self.TR_array

        if show_networks:
            if "nodes_info" in self.TS_info:
                node_networks = node_info2network(self.TS_info["nodes_info"])
            elif "node_labels" in self.TS_info:
                node_networks = node_labels2networks(self.TS_info["node_labels"])
        else:
            node_networks = None

        if rank_norm:
            dFC_dict = rank_norm_dFC_dict(self.dFC2dict(TRs=TRs))
            cmap = "plasma"
            center_0 = False
        else:
            dFC_dict = self.dFC2dict(TRs=TRs)
            cmap = "seismic"
            center_0 = True

        visualize_conn_mat_dict(
            data=dFC_dict,
            title=self.measure.measure_name + " dFC",
            fix_lim=fix_lim,
            normalize=normalize,
            node_networks=node_networks,
            cmap=cmap,
            center_0=center_0,
            save_image=save_image,
            output_root=output_root,
        )



"""
Time Series class

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import os
import warnings
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from scipy import signal

from .dfc_utils import print_dict

################################# Parameters ####################################

## visualization parameters
fig_dpi = 120
fig_bbox_inches = "tight"
fig_pad = 0.1
show_title = False
save_fig_format = "png"

################################# TIME_SERIES class ######################################

"""

todo:
- select nodes for visualizer
- default node list is chosen by arange !
"""


class TIME_SERIES:
    def __init__(
        self,
        data=None,
        subj_id=None,
        Fs=None,
        time_array=None,
        locs=None,
        node_labels=None,
        TS_name="",
        session_name="",
    ):
        """
        subj_id is an id to identify the subjects
        all properties are applied to every subject separately

        It is recommended that the time_array of all subjects is equal
        """

        assert (
            (not data is None) and (not Fs is None) and (not subj_id is None)
        ), "data, subj_id, and Fs args must be provided."

        assert type(locs) is np.ndarray, "locs must be a numpy array"
        assert type(node_labels) is list, "node_labels must be a list"
        assert locs.shape[0] == len(
            node_labels
        ), "locs and node_labels must have the same length"
        assert locs.shape[1] == 3, "locs must have 3 columns"

        self.data_dict_ = {}
        self.data_dict_[subj_id] = {}
        self.data_dict_[subj_id]["data"] = data
        self.data_ = None
        self.Fs_ = Fs
        self.Fs_ratio_ = 1.00
        self.noise_ratio = 0.0
        self.normalized = False
        self.TS_name_ = TS_name
        self.session_name_ = session_name
        self.n_regions_ = data.shape[0]

        if time_array is None:
            self.data_dict_[subj_id]["time_array"] = 1 / self.Fs_ + np.arange(
                0, data.shape[1] / self.Fs_, 1 / self.Fs_
            )
        else:
            self.data_dict_[subj_id]["time_array"] = time_array

        self.locs_ = locs
        self.node_labels_ = node_labels

        self.nodes_selection_ = list(range(self.n_regions_))

    @property
    def info(self):
        print_dict(self.info_dict)

    @property
    def info_dict(self):
        info_dict = {}
        info_dict["n_time"] = self.n_time
        info_dict["n_regions"] = self.n_regions
        info_dict["Fs"] = self.Fs
        info_dict["Fs_ratio"] = self.Fs_ratio_
        info_dict["noise_ratio"] = self.noise_ratio
        info_dict["nodes_lst"] = self.nodes_lst
        info_dict["node_labels"] = self.node_labels
        info_dict["nodes_locs"] = self.locs
        info_dict["subj_id_lst"] = self.subj_id_lst
        info_dict["time"] = self.time

        return info_dict

    @property
    def data_dict(self):
        return self.data_dict_

    @property
    def data(self):
        if self.data_ is None:
            self.update_data()
        return self.data_

    def update_data(self):
        # after any change in data_dict, self.data_ is
        # set to None and needs an update before being used
        data = None
        for subj in self.data_dict:
            if data is None:
                data = self.data_dict[subj]["data"]
            else:
                data = np.concatenate((data, self.data_dict[subj]["data"]), axis=1)
        self.data_ = data
        return

    @property
    def subj_id_lst(self):
        return [subj_id for subj_id in self.data_dict]

    @property
    def nodes_lst(self):
        # output shape is (n_region,)
        return np.array(self.nodes_selection_)

    @property
    def locs(self):
        """
        locs shape is (n_region, 3)
        """
        if self.locs_ is None:
            return None
        else:
            return self.locs_[self.nodes_lst, :]

    @property
    def node_labels(self):
        if self.node_labels_ is None:
            return None
        else:
            return [self.node_labels_[i] for i in self.nodes_lst]

    @property
    def Fs(self):
        return self.Fs_ * self.Fs_ratio_

    @property
    def n_time(self):
        # if there are multiple subjects, the time_array is ambiguous, so None is returned
        if len(self.subj_id_lst) > 1:
            return None
        else:
            return len(self.time)

    @property
    def n_regions(self):
        return len(self.nodes_lst)

    @property
    def time(self):
        # if there are multiple subjects, the time_array is ambiguous, so None is returned
        if len(self.subj_id_lst) > 1:
            return None
        else:
            return self.data_dict[self.subj_id_lst[0]]["time_array"]

    @property
    def TS_name(self):
        return self.TS_name_

    def get_subj_ts(self, subjs_id=None):
        """
        you can select time samples by their subj_id
        ! be careful about the original properties of TS hidden in new TS
        node selection will be kept.

        subjs_id is one str or a list of str
        """

        flag = 0
        if not type(subjs_id) is list:
            subjs_id = [subjs_id]
            flag = 1

        # the reason we don't make it from scratch is that we want to
        # keep the node selection and manipulations
        new_TS = deepcopy(self)

        SUBJECTS = [subj_id for subj_id in new_TS.data_dict_]
        for subj in SUBJECTS:
            if not subj in subjs_id:
                new_TS.data_dict_.pop(subj, None)

        if flag == 1:
            new_TS.TS_name_ = self.TS_name + " subject " + subjs_id[0]

        return new_TS

    def append_ts(self, new_time_series, time_array=None, subj_id=None):
        # append new time series numpy array to existing ones
        # truncate and node selection , etc will be automatically applied to new TS;
        # However, at first the new TS must have the same properties as the original properties of
        # the existing TSs

        assert self.n_regions_ == new_time_series.shape[0], "Number of nodes mismatch."

        assert not subj_id is None, "subj_id must be provided."

        assert not subj_id in self.data_dict_, "subj_id already exists."

        self.data_dict_[subj_id] = {}

        if not time_array is None:
            self.data_dict_[subj_id]["time_array"] = time_array
        else:
            self.data_dict_[subj_id]["time_array"] = 1 / self.Fs_ + np.arange(
                0, new_time_series.shape[1] / self.Fs_, 1 / self.Fs_
            )

        self.data_dict_[subj_id]["data"] = new_time_series

        self.data_ = None

    def concat_ts(self, new_TS):
        """
        concatenate another Time Series obj
        to the current one.
        It assumes that the new TS has only one subject
        """
        assert self.Fs == new_TS.Fs, "Fs mismatch!"
        assert len(self.node_labels) == len(new_TS.node_labels), "node_labels mismatch!"
        for i in range(len(self.node_labels)):
            assert self.node_labels[i] == new_TS.node_labels[i], "node_labels mismatch!"
            assert np.all(self.locs[i, :] == new_TS.locs[i, :]), "locs mismatch!"
        assert len(new_TS.subj_id_lst) == 1, "new_TS must have only one subject."

        self.append_ts(new_time_series=new_TS.data, subj_id=new_TS.subj_id_lst[0])

    def truncate(self, start_time=None, end_time=None, start_point=None, end_point=None):

        # truncates TS of every subj separately
        # based on either time or samples
        # if all None -> whole time_series

        for subj_id in self.data_dict:

            time_array = self.data_dict[subj_id]["time_array"]
            n_time = self.data_dict[subj_id]["data"].shape[1]
            assert n_time == len(time_array), "time_array and data mismatch."

            start = 0
            end = n_time

            if not start_point is None:
                start = start_point

            if not end_point is None:
                end = end_point + 1

            if not start_time is None:
                start = np.argwhere(time_array >= start_time)[0, 0]

            if not end_time is None:
                end = np.argwhere(time_array <= end_time)[-1, 0] + 1

            start = max(start, 0)
            end = min(end, n_time)
            assert start < end, "start must be less than end."

            self.data_dict_[subj_id]["data"] = self.data_dict[subj_id]["data"][
                :, start:end
            ]
            self.data_dict_[subj_id]["time_array"] = time_array[start:end]

            self.data_ = None

    def normalize(self):
        # normalization
        if not self.normalized:
            for subj_id in self.data_dict:
                new_time_series = self.data_dict[subj_id]["data"]
                for n in range(new_time_series.shape[0]):
                    new_time_series[n, :] = new_time_series[n, :] - np.mean(
                        new_time_series[n, :]
                    )
                    new_time_series[n, :] = np.divide(
                        new_time_series[n, :], np.std(new_time_series[n, :])
                    )
                self.data_dict_[subj_id]["data"] = new_time_series

            self.normalized = True
            self.data_ = None

    def spatial_downsample(self, num_select_nodes, rand_node_slct=False):
        if num_select_nodes < self.n_regions:
            if rand_node_slct:
                np.random.seed(0)
                nodes_idx = np.random.choice(
                    range(self.n_regions), size=num_select_nodes, replace=False
                )
                nodes_idx.sort()
            else:
                # nodes_idx = np.array(list(range(self.num_select_nodes)))
                nodes_idx = np.arange(
                    0,
                    self.n_regions,
                    np.ceil(self.n_regions / num_select_nodes),
                    dtype=int,
                )
            self.select_nodes(nodes_idx=nodes_idx)

            self.data_ = None

    def Fs_resample(self, Fs_ratio=None):
        # downsample frequency
        if Fs_ratio != 1 and self.Fs_ratio_ == 1:
            for subj_id in self.data_dict:
                new_time_series = self.data_dict[subj_id]["data"]
                downsampled_time_series = np.zeros(
                    (new_time_series.shape[0], int(new_time_series.shape[1] * Fs_ratio))
                )
                for n in range(new_time_series.shape[0]):
                    downsampled_time_series[n, :] = signal.resample(
                        new_time_series[n, :], int(new_time_series.shape[1] * Fs_ratio)
                    )
                self.data_dict_[subj_id]["data"] = downsampled_time_series

                # find the new time_array after resampling
                _, new_time_array = signal.resample(
                    new_time_series[n, :],
                    int(new_time_series.shape[1] * Fs_ratio),
                    t=self.data_dict[subj_id]["time_array"],
                )
                self.data_dict_[subj_id]["time_array"] = new_time_array

            self.Fs_ratio_ = Fs_ratio
            self.data_ = None

    def add_noise(self, noise_ratio, mean_noise=0):
        # adding noise perturbation
        if noise_ratio > 0 and self.noise_ratio == 0:
            for subj_id in self.data_dict:
                new_time_series = self.data_dict[subj_id]["data"]
                power_signal = np.mean(new_time_series**2)
                power_noise = power_signal * noise_ratio
                new_time_series += np.random.normal(
                    mean_noise,
                    np.sqrt(power_noise),
                    (new_time_series.shape[0], new_time_series.shape[1]),
                )
                self.data_dict_[subj_id]["data"] = new_time_series

            self.noise_ratio = noise_ratio
            self.data_ = None

    def select_subjs(self, num_subj):
        # selects the first num_subj subjects in self.subj_id_lst

        SUBJECTS = self.subj_id_lst
        if num_subj < len(SUBJECTS):
            for subj_id in SUBJECTS:
                if not subj_id in SUBJECTS[:num_subj]:
                    self.data_dict_.pop(subj_id, None)

            self.data_ = None

    def select_nodes(self, nodes_idx=None):
        # select the nodes indexed by numbers in nodes_idx. nodes_idx is a numpy 1D array
        # if nodes_idx is None -> all the nodes will be considered (resets node selection)
        # if nodes_idx is not sorted, it can be used to reorder the nodes
        # this function can be used only once (you cannot select the nodes again)

        if nodes_idx is None:
            self.nodes_selection_ = np.arange(0, self.n_regions_, 1, dtype=int)
        else:
            self.nodes_selection_ = nodes_idx

        for subj in self.data_dict_:
            self.data_dict_[subj]["data"] = self.data_dict_[subj]["data"][
                self.nodes_lst, :
            ]

        self.data_ = None

    def visualize(
        self,
        start_time=None,
        end_time=None,
        nodes_lst=None,
        save_image=False,
        output_root=None,
    ):
        """
        time in seconds
        nodes_lst is a list of indices
        """

        if len(self.subj_id_lst) > 1:
            # raise a user warning
            warnings.warn(
                "Multiple subjects are not supported in visualization.", UserWarning
            )
            return

        start = 0
        end = self.n_time

        if not start_time is None:
            start = np.argwhere(self.time >= start_time)[0, 0]

        if not end_time is None:
            end = np.argwhere(self.time <= end_time)[-1, 0] + 1

        interval = list(range(start, end))

        if nodes_lst is None:
            nodes_lst = self.nodes_lst[:, np.newaxis]
        else:
            nodes_lst = np.array(nodes_lst)[:, np.newaxis]

        plt.figure(figsize=(15, 5))
        plt.plot(self.time[interval], self.data[nodes_lst, interval].T)
        plt.xlabel("time (sec)")

        if not self.session_name_ is None:
            session_name = self.session_name_
        else:
            session_name = ""

        title = self.TS_name_ + " " + session_name
        if show_title:
            plt.title(title)
        if save_image:
            folder = output_root[: output_root.rfind("/")]
            file_name = title.replace(" ", "_")
            if not os.path.exists(folder):
                os.makedirs(folder)
            plt.savefig(
                output_root + file_name + "." + save_fig_format,
                dpi=fig_dpi,
                bbox_inches=fig_bbox_inches,
                pad_inches=fig_pad,
                format=save_fig_format,
            )
            plt.close()
        else:
            plt.show()



import json
import os
import warnings

import numpy as np

from pydfc import data_loader, task_utils

warnings.simplefilter("ignore")

################################# Parameters #################################
# data paths
# main_root = '../../DATA/ds002785' # for local
main_root = "../../../DATA/task-based/openneuro/ds002785"  # for server
fmriprep_root = f"{main_root}/derivatives/fmriprep"
output_root = f"{main_root}/derivatives/ROI_timeseries"

bold_suffix = "_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"

# for consistency we use 0 for resting state
TASKS = [
    "task-restingstate",
    "task-anticipation",
    "task-emomatching",
    "task-faces",
    "task-gstroop",
    "task-workingmemory",
]

# find all subjects
ALL_SUBJs = os.listdir(fmriprep_root)
ALL_SUBJs = [i for i in ALL_SUBJs if ("sub-" in i) and (not ".html" in i)]
ALL_SUBJs.sort()

# pick the subject
job_id = int(os.getenv("SGE_TASK_ID"))
subj = ALL_SUBJs[job_id - 1]  # SGE_TASK_ID starts from 1 not 0

print(
    f"subject-level ROI signal extraction CODE started running ... for subject: {subj} ..."
)
################################# FIND THE FUNC FILE #################################
for task in TASKS:
    # find the func file for this subject and task
    ALL_TASK_FILES = os.listdir(f"{fmriprep_root}/{subj}/func/")
    ALL_TASK_FILES = [
        i for i in ALL_TASK_FILES if (bold_suffix in i) and (task in i)
    ]  # only keep the denoised files? or use the original files?
    # print(ALL_TASK_FILES)
    if not len(ALL_TASK_FILES) == 1:
        # if the func file is not found, exclude the subject
        print("Func file not found for " + subj + " " + task)
        continue
    fmriprep_file = f"{fmriprep_root}/{subj}/func/{ALL_TASK_FILES[0]}"
    info_file = (
        f"{main_root}/{subj}/func/{ALL_TASK_FILES[0].replace(bold_suffix, '_bold.json')}"
    )

    ################################# LOAD JSON INFO #########################
    # Opening JSON file as a dictionary
    f = open(info_file)
    acquisition_data = json.load(f)
    f.close()
    TR_mri = acquisition_data["RepetitionTime"]
    ################################# EXTRACT TIME SERIES #########################
    # extract ROI signals and convert to TIME_SERIES object
    time_series = data_loader.nifti2timeseries(
        nifti_file=fmriprep_file,
        n_rois=100,
        Fs=1 / TR_mri,
        subj_id=subj,
        confound_strategy="no_motion",
        standardize="zscore",
        TS_name="BOLD",
        session=task,
    )
    num_time_mri = time_series.n_time
    ################################# EXTRACT TASK LABELS #########################
    oversampling = 50  # more samples per TR than the func data to have a better event_labels time resolution
    if task == "task-restingstate":
        events = []
        event_types = ["rest"]
        event_labels = np.zeros((int(num_time_mri * oversampling), 1))
        task_labels = np.zeros((int(num_time_mri * oversampling), 1))
        Fs_task = float(1 / TR_mri) * oversampling
    else:
        task_events_root = f"{main_root}/{subj}/func/"
        ALL_EVENTS_FILES = os.listdir(task_events_root)
        ALL_EVENTS_FILES = [
            i
            for i in ALL_EVENTS_FILES
            if (subj in i) and (task in i) and ("events.tsv" in i)
        ]
        if not len(ALL_EVENTS_FILES) == 1:
            # if the events file is not found, exclude the subject
            print("Events file not found for " + subj + " " + task)
            continue
        # load the tsv events file
        events_file = task_events_root + ALL_EVENTS_FILES[0]
        events = np.genfromtxt(events_file, delimiter="\t", dtype=str)
        # get the task labels
        event_types = ["rest"] + list(np.unique(events[1:, 2]))
        event_labels, Fs_task = task_utils.events_time_to_labels(
            events=events,
            TR_mri=TR_mri,
            num_time_mri=num_time_mri,
            event_types=event_types,
            oversampling=oversampling,
            return_0_1=False,
        )
        # fill task labels with 0 (rest) and k (task's index)
        task_labels = np.multiply(event_labels != 0, TASKS.index(task))
    ################################# SAVE #################################
    # save the ROI time series and task data
    task_data = {
        "task": task,
        "task_labels": task_labels,
        "task_types": TASKS,
        "event_labels": event_labels,
        "event_types": event_types,
        "events": events,
        "Fs_task": Fs_task,
        "TR_mri": TR_mri,
        "num_time_mri": num_time_mri,
    }
    subj_folder = f"{subj}_{task}"
    if not os.path.exists(f"{output_root}/{subj_folder}/"):
        os.makedirs(f"{output_root}/{subj_folder}/")
    np.save(f"{output_root}/{subj_folder}/time_series.npy", time_series)
    np.save(f"{output_root}/{subj_folder}/task_data.npy", task_data)

print(
    f"subject-level ROI signal extraction CODE finished running ... for subject: {subj} ..."
)
####################################################################



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 8 2023

@author: mte
"""

import math
import os
import sys
import warnings
from copy import deepcopy
from math import ceil

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as shc
import scipy.spatial.distance as ssd
import seaborn as sns
import statsmodels.api as sm
from matplotlib.colors import ListedColormap
from nilearn.plotting import plot_markers
from sklearn.manifold import TSNE
from statsmodels.formula.api import ols

sys.path.append("./../git_codes/BIC_codes/")
from functions.dFC_funcs import (
    dFC_mat2vec,
    mat_reorder,
    unzip_name,
    visualize_conn_mat_dict,
    zip_name,
)

################################# Parameters ####################################

fig_dpi = 120
fig_bbox_inches = "tight"
fig_pad = 0.1
show_title = False
save_fig_format = "png"  # pdf, png,

################################# Plotting Functions ####################################


def title2file_name(title):
    """
    change all spaces in the title to _
    the original string remains unchanged
    """
    return title.replace(" ", "_")


def plot_sample_dFC(
    D,
    x,
    title="",
    cmap="seismic",
    normalize=False,
    disp_diag=True,
    save_image=False,
    output_root=None,
    fix_lim=True,
    center_0=True,
    node_networks=None,
    segmented=False,
):
    """
    D is a dictionary of dFC samples. each
    key is the name of a dFC matrix (e.g. method
    used for assessing it), and D[key][x] contains the
    the dFC matrix as a numpy ndarray
    """

    num_dFC = len(D)
    names_lst = [key for key in D]
    num_time = len(D[names_lst[0]][x])

    fig_width = 48 * (num_time / 10)
    fig_height = 55 * (num_dFC / 10)

    fig, axes = plt.subplots(
        num_dFC, num_time, figsize=(fig_width, fig_height), facecolor="w", edgecolor="k"
    )

    fig.subplots_adjust(bottom=0.1, top=0.85, left=0.1, right=0.9, wspace=0.5, hspace=0.6)

    for i, dFC_mat_name in enumerate(D):
        visualize_conn_mat_dict(
            data=D[dFC_mat_name][x],
            node_networks=node_networks,
            title=dFC_mat_name,
            cmap=cmap,
            center_0=center_0,
            normalize=normalize,
            fix_lim=fix_lim,
            disp_diag=disp_diag,
            segmented=segmented,
            save_image=False,
            output_root=output_root,
            axes=axes[i, :],
            fig=fig,
        )

    fig.subplots_adjust(bottom=0.1, top=0.85, left=0.1, right=0.9, wspace=0.5, hspace=0.6)

    # set row names
    for i, dFC_mat_name in enumerate(D):
        axes[i, 0].set_ylabel(
            dFC_mat_name, fontdict={"fontsize": 25, "fontweight": "bold"}, rotation=90
        )

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        fig.savefig(
            output_root + title.replace(" ", "_") + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def plot_rois(node_networks, nodes_locs, save_image=False, output_root=None):

    networks = list(np.unique(node_networks))

    fig_width = 25
    fig_height = len(networks)

    fig, axes = plt.subplots(
        ceil(len(networks) / 3),
        3,
        figsize=(fig_width, fig_height),
        facecolor="w",
        edgecolor="k",
    )

    axes = axes.ravel()

    fig.subplots_adjust(
        bottom=0.1, top=0.85, left=0.1, right=0.9, wspace=0.03, hspace=0.3
    )

    for i, target_network in enumerate(networks):

        locs = []
        node_values = []
        for node_id, node_network in enumerate(node_networks):
            if node_network == target_network:
                node_values.append(1)
                locs.append(nodes_locs[node_id])

        node_values = np.array(node_values)
        locs = np.array(locs)

        plot_markers(
            node_values=node_values,
            node_coords=locs,
            node_size=100,
            node_cmap="Reds",
            node_vmax=1,
            node_vmin=0,
            annotate=True,
            colorbar=False,
            axes=axes[i],
        )

    title = "Resting State Networks"
    # set subplot titles
    for i, network in enumerate(networks):
        axes[i].title.set_text(f"{network} network")
        axes[i].title.set_size(20)
        axes[i].title.set_weight("bold")

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        fig.savefig(
            output_root + title.replace(" ", "_") + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def pairwise_cat_plots(
    data=None,
    x=None,
    y=None,
    z=None,
    title="",
    label_dict={},
    save_image=False,
    output_root=None,
):
    """
    data is a dictionary with different vars as keys
    if z is specidied, it will be used as a out of distribution
    sample, e.g. actual similarity when plotting randomized
    distribution.
    """

    sns.set_context("paper", font_scale=2.5, rc={"lines.linewidth": 3.0})

    row_keys = [key for key in data]
    n_rows = len(row_keys)
    column_keys = [key for key in data[row_keys[-1]]]
    n_columns = len(column_keys)

    sns.set_style("darkgrid")

    fig_width = n_columns * 5
    fig_height = n_rows * 5
    fig, axs = plt.subplots(
        n_rows,
        n_columns,
        figsize=(fig_width, fig_height),
        facecolor="w",
        edgecolor="k",
        sharex=True,
        sharey=True,
    )

    axs_plotted = list()
    for i, key_i in enumerate(data):
        for j, key_j in enumerate(data[key_i]):
            df = pd.DataFrame(data[key_i][key_j])

            if not z is None:
                sns.stripplot(
                    ax=axs[i, j], data=df, x=x, y=z, color="red", jitter=False, size=10
                )
            sns.violinplot(ax=axs[i, j], data=df, x=x, y=y)

            axs[i, j].set_title(
                key_i + "-" + key_j, fontdict={"fontsize": 25, "fontweight": "bold"}
            )

            ## set labels
            ylabel = axs[i, j].get_ylabel()
            if ylabel in label_dict:
                ylabel = label_dict[ylabel]
            axs[i, j].set_ylabel(ylabel, fontdict={"fontsize": 20, "fontweight": "bold"})
            xlabel = axs[i, j].get_xlabel()
            if xlabel in label_dict:
                xlabel = label_dict[xlabel]
            axs[i, j].set_xlabel(xlabel, fontdict={"fontsize": 20, "fontweight": "bold"})
            # set font size of the tick labels and make them bold
            tick_labels = axs[i, j].get_xticklabels() + axs[i, j].get_yticklabels()
            for label in tick_labels:
                label.set_fontweight("bold")

            axs_plotted.append(axs[i, j])

    # remove extra subplots
    for ax in axs.ravel():
        if not ax in axs_plotted:
            ax.set_axis_off()
            ax.xaxis.set_tick_params(which="both", labelbottom=True)

    if show_title:
        plt.suptitle(title, fontsize=15, y=0.90)

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def joint_dist_plot(data, title="", label_dict={}, save_image=False, output_root=None):
    """
    data is a dictionary including list of dFC values
    of each dFC method
    """
    df = pd.DataFrame(data)
    fig_width = 5 * len(data)
    fig_height = 5 * len(data)

    sns.set_context("paper", font_scale=2.5, rc={"lines.linewidth": 3.0})

    sns.set_style("darkgrid")

    g = sns.PairGrid(df)

    g.map_diag(sns.histplot)
    g.map_offdiag(sns.histplot)

    g.fig.set_figwidth(fig_width)
    g.fig.set_figheight(fig_height)
    g.fig.subplots_adjust(top=0.95)

    for i in range(g.axes.shape[0]):
        for j in range(g.axes.shape[1]):

            ## set labels
            ylabel = g.axes[i, j].get_ylabel()
            if ylabel in label_dict:
                ylabel = label_dict[ylabel]
            g.axes[i, j].set_ylabel(
                ylabel, fontdict={"fontsize": 25, "fontweight": "bold"}
            )
            xlabel = g.axes[i, j].get_xlabel()
            if xlabel in label_dict:
                xlabel = label_dict[xlabel]
            g.axes[i, j].set_xlabel(
                xlabel, fontdict={"fontsize": 25, "fontweight": "bold"}
            )

            # set font size of the tick labels and make them bold
            tick_labels = g.axes[i, j].get_xticklabels() + g.axes[i, j].get_yticklabels()
            for label in tick_labels:
                label.set_fontweight("bold")

    if show_title:
        plt.suptitle(title, fontsize=50, y=0.98)

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def pairwise_scatter_plots(
    data,
    x,
    y,
    title="",
    hist=False,
    label_dict={},
    equal_axis_lim=False,
    show_x_equal_y=False,
    save_image=False,
    output_root=None,
):
    """
    data is a dictionary with different vars as keys
    """

    sns.set_context("paper", font_scale=2.5, rc={"lines.linewidth": 3.0})

    row_keys = [key for key in data]
    n_rows = len(row_keys)
    column_keys = [key for key in data[row_keys[-1]]]
    n_columns = len(column_keys)

    sns.set_style("darkgrid")

    fig_width = n_columns * 5
    fig_height = n_rows * 5
    fig, axs = plt.subplots(
        n_rows,
        n_columns,
        figsize=(fig_width, fig_height),
        facecolor="w",
        edgecolor="k",
        sharex=True,
        sharey=True,
    )

    # equal x_lim and y_lim
    if equal_axis_lim or show_x_equal_y:
        min_lim = None
        max_lim = None
        for i, key_i in enumerate(data):
            for j, key_j in enumerate(data[key_i]):
                df = pd.DataFrame(data[key_i][key_j])
                m = np.minimum(df[x].min(), df[y].min())
                M = np.maximum(df[x].max(), df[y].max())
                if min_lim is None:
                    min_lim = m
                    max_lim = M
                else:
                    min_lim = np.minimum(m, min_lim)
                    max_lim = np.maximum(M, max_lim)

        lim_L = max_lim - min_lim
        min_lim = min_lim - lim_L * 0.1
        max_lim = max_lim + lim_L * 0.1

    axs_plotted = list()
    for i, key_i in enumerate(data):
        for j, key_j in enumerate(data[key_i]):
            df = pd.DataFrame(data[key_i][key_j])
            if hist:
                g = sns.histplot(ax=axs[i, j], data=df, x=x, y=y, bins=50)
            else:
                g = sns.scatterplot(ax=axs[i, j], data=df, x=x, y=y, s=50)
            axs[i, j].set_title(
                key_i + "-" + key_j, fontdict={"fontsize": 25, "fontweight": "bold"}
            )

            ## set labels and font sizes
            ylabel = g.get_ylabel()
            if ylabel in label_dict:
                ylabel = label_dict[ylabel]
            g.set_ylabel(ylabel, fontdict={"fontsize": 18, "fontweight": "bold"})
            xlabel = g.get_xlabel()
            if xlabel in label_dict:
                xlabel = label_dict[xlabel]
            g.set_xlabel(xlabel, fontdict={"fontsize": 18, "fontweight": "bold"})
            g.tick_params(axis="x", which="major", labelsize=18)
            g.tick_params(axis="y", which="major", labelsize=18)
            tick_labels = g.get_xticklabels() + g.get_yticklabels()
            for label in tick_labels:
                label.set_fontweight("bold")

            # equal x_lim and y_lim
            if equal_axis_lim:
                axs[i, j].set_xlim(min_lim, max_lim)
                axs[i, j].set_ylim(min_lim, max_lim)

            # y=x line
            if show_x_equal_y:
                X_plot = np.linspace(min_lim, max_lim, 100)
                Y_plot = X_plot
                axs[i, j].plot(X_plot, Y_plot, color="r")

            axs_plotted.append(axs[i, j])
    # remove extra subplots
    for ax in axs.ravel():
        if not ax in axs_plotted:
            ax.set_axis_off()
            ax.xaxis.set_tick_params(which="both", labelbottom=True)

    if show_title:
        plt.suptitle(title, fontsize=15, y=0.90)

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def scatter_plot(
    data,
    x,
    y,
    labels=None,
    hue=None,
    title="",
    hist=False,
    label_dict={},
    equal_axis_lim=False,
    show_x_equal_y=False,
    c=0.25,
    save_image=False,
    output_root=None,
):
    """
    data is a dictionary with different vars as keys
    c determines how far the annotation will be from dots
    """
    df = pd.DataFrame(data)

    sns.set_context("paper", font_scale=2.5, rc={"lines.linewidth": 3.0})

    fig_width = 20
    fig_height = 20
    plt.figure(figsize=(fig_width, fig_height))
    sns.set_style("darkgrid")
    if hist:
        g = sns.histplot(data=df, x=x, y=y, hue=hue)
    else:
        g = sns.scatterplot(data=df, x=x, y=y, s=100, hue=hue)

    ## set labels and font sizes
    ylabel = g.get_ylabel()
    if ylabel in label_dict:
        ylabel = label_dict[ylabel]
    g.set_ylabel(ylabel, fontdict={"fontsize": 35, "fontweight": "bold"})
    xlabel = g.get_xlabel()
    if xlabel in label_dict:
        xlabel = label_dict[xlabel]
    g.set_xlabel(xlabel, fontdict={"fontsize": 35, "fontweight": "bold"})
    g.tick_params(axis="x", which="major", labelsize=30)
    g.tick_params(axis="y", which="major", labelsize=30)
    tick_labels = g.get_xticklabels() + g.get_yticklabels()
    for label in tick_labels:
        label.set_fontweight("bold")

    # equal x_lim and y_lim
    if equal_axis_lim:
        min_lim = np.minimum(df[x].min(), df[y].min())
        max_lim = np.maximum(df[x].max(), df[y].max())
        lim_L = max_lim - min_lim
        min_lim = min_lim - lim_L * 0.1
        max_lim = max_lim + lim_L * 0.1
        g.set_xlim(min_lim, max_lim)
        g.set_ylim(min_lim, max_lim)

    # y=x line
    if show_x_equal_y:
        min_lim = np.minimum(df[x].min(), df[y].min())
        max_lim = np.maximum(df[x].max(), df[y].max())
        lim_L = max_lim - min_lim
        min_lim = min_lim - lim_L * 0.1
        max_lim = max_lim + lim_L * 0.1
        X_plot = np.linspace(min_lim, max_lim, 100)
        Y_plot = X_plot
        plt.plot(X_plot, Y_plot, color="r")

    if (not labels is None) and (not hist):
        # the labels are located smartly
        # the direction will be away from mean
        # the distance will be inverse proportional to
        # distance from mean
        mid_x = (np.max(df[x]) + np.min(df[x])) / 2
        mid_y = (np.max(df[y]) + np.min(df[y])) / 2
        x_range = max(np.max(df[x]) - mid_x, mid_x - np.min(df[x]))
        y_range = max(np.max(df[y]) - mid_y, mid_y - np.min(df[y]))
        distance_from_mean_range = math.sqrt(x_range**2 + y_range**2)
        for i in range(len(df[x])):
            distance_from_mean = math.sqrt(
                (df[x][i] - mid_x) ** 2 + (df[y][i] - mid_y) ** 2
            )
            text_x = (
                df[x][i]
                + c
                * np.sign(df[x][i] - mid_x)
                * np.abs(df[x][i] - mid_x)
                * (distance_from_mean_range - distance_from_mean)
                / distance_from_mean
            )
            text_y = (
                df[y][i]
                + c
                * np.sign(df[y][i] - mid_y)
                * np.abs(df[y][i] - mid_y)
                * (distance_from_mean_range - distance_from_mean)
                / distance_from_mean
            )
            plt.text(
                x=text_x,
                y=text_y,
                s=df[labels][i],
                fontdict=dict(color="black", size=14, weight="bold"),
            )
            plt.plot([df[x][i], text_x], [df[y][i], text_y], "k", linewidth=0.5)

    if show_title:
        plt.title(title, fontsize=15)

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def cat_plot(
    data,
    x,
    y,
    kind="bar",
    scale_dist=False,
    log=False,
    title="",
    label_dict={},
    y_lim=None,
    save_image=False,
    output_root=None,
):
    """
    data is a dictionary with different vars as keys
    kind can be = box or violin or bar
    scale_dist is only for kind=='violin'
    """

    sns.set_context("paper", font_scale=1.0, rc={"lines.linewidth": 1.0})

    sns.set_style("darkgrid")

    df = pd.DataFrame(data)

    fig_width = 2 * len(np.unique(data[x]))
    fig_height = 5

    if kind == "violin" and scale_dist:
        g = sns.catplot(
            data=df,
            x=x,
            y=y,
            kind=kind,
            scale="width",
            # errorbar=("pi", 95)
        )
    elif kind == "bar":
        g = sns.catplot(
            data=df,
            x=x,
            y=y,
            kind=kind,
            width=0.25,
            # errorbar=("pi", 95)
        )
    elif kind == "box":
        g = sns.catplot(
            data=df,
            x=x,
            y=y,
            kind=kind,
            width=0.25,
            fliersize=1.0,
            # errorbar=("pi", 95)
        )
    else:
        g = sns.catplot(
            data=df,
            x=x,
            y=y,
            kind=kind,
            # errorbar=("pi", 95)
        )

    if log:
        plt.yscale("log")

    g.fig.set_figwidth(fig_width)
    g.fig.set_figheight(fig_height)

    ## set labels
    ylabel = g.ax.get_ylabel()
    if ylabel in label_dict:
        ylabel = label_dict[ylabel]
    g.ax.set_ylabel(ylabel, fontdict={"fontsize": 13, "fontweight": "bold"})
    xlabel = g.ax.get_xlabel()
    if xlabel in label_dict:
        xlabel = label_dict[xlabel]
    g.ax.set_xlabel(xlabel, fontdict={"fontsize": 13, "fontweight": "bold"})
    # set font size of the tick labels and make them bold
    tick_labels = g.ax.get_xticklabels() + g.ax.get_yticklabels()
    for label in tick_labels:
        label.set_fontweight("bold")
    if not y_lim is None:
        g.ax.set_ylim(y_lim)

    if show_title:
        plt.title(title, fontsize=15)
    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def visualize_sim_mat(
    data,
    mat_key,
    title="",
    name_lst_key=None,
    cmap="viridis",
    annot=True,
    fmt=2,
    label_dict={},
    show_diag=False,
    show_sig=False,
    no_color=False,
    save_image=False,
    output_root=None,
    axes=None,
    fig=None,
):
    """
    - name_lst_key is the key to list of names
    - data must be a dict of correlation/connectivity matrices
    - masks the nan values
    sample:
    Suptitle1
        corr_mat
            0.00 0.31 0.76
            0.31 0.00 0.43
            0.76 0.43 0.00
        measure_lst
            ContinuousHMM
            Windowless
            Clustering_pear_corr
    Suptitle1
        corr_mat
            0.00 0.32 0.76
            0.32 0.00 0.45
            0.76 0.45 0.00
        measure_lst
            ContinuousHMM
            Windowless
            Clustering_pear_corr
    """

    sns.set_context("paper", font_scale=1.0, rc={"lines.linewidth": 1.0})

    sns.set_style("white")

    if no_color:
        cmap = ListedColormap(["white"])

    if name_lst_key is None:
        fig_width = int(25 * (len(data) / 10))
    else:
        fig_width = int(60 * (len(data) / 10) + 1)
    fig_height = 5

    fig_flag = True
    if axes is None or fig is None:
        fig_flag = False

    if not fig_flag:
        fig, axes = plt.subplots(
            1,
            len(data),
            figsize=(fig_width, fig_height),
            facecolor="w",
            edgecolor="k",
            sharey=False,
        )

    if not type(axes) is np.ndarray:
        axes = np.array([axes])

    if show_title:
        fig.suptitle(title, fontsize=20, y=0.98)  # , fontsize=20, size=20

    axes = axes.ravel()

    # normalizing and scale
    sim_mats = list()
    for i, key in enumerate(data):
        sim_mats.append(data[key][mat_key])
    sim_mats = np.array(sim_mats)

    # plot
    for i, key in enumerate(data):

        C = sim_mats[i, :, :]

        name_lst = None
        if not name_lst_key is None:
            name_lst = data[key][name_lst_key]

        cbar_flag = False
        # if i==(len(data)-1):
        #     cbar_flag = True

        if annot:
            C_forlabels = C.copy()
            if not show_diag:
                np.fill_diagonal(C_forlabels, np.nan)
            df = pd.DataFrame(C_forlabels)
            if show_sig:
                annot_labels = df.applymap(
                    lambda v: (
                        ""
                        if np.isnan(v)
                        else str(round(v, fmt))
                        + "".join(["*" for t in [0.05, 0.01, 0.001] if v <= t])
                    )
                )
            else:
                annot_labels = df.applymap(
                    lambda v: "" if np.isnan(v) else str(round(v, fmt))
                )
        else:
            annot_labels = False

        # borderlines color
        if no_color:
            linecolor = "black"
            annot_kws = {"weight": "bold"}
        else:
            linecolor = "w"
            annot_kws = {"weight": "bold"}

        im = sns.heatmap(
            C,
            annot=annot_labels,
            annot_kws=annot_kws,
            fmt="",
            cmap=cmap,
            xticklabels=name_lst,
            yticklabels=name_lst,
            ax=axes[i],
            cbar=cbar_flag,
            square=True,
            linewidth=2,
            linecolor=linecolor,
        )
        axis_title = key
        if key in label_dict:
            axis_title = label_dict[key]
        axes[i].set_title(axis_title, fontdict={"fontsize": 18, "fontweight": "bold"})
        im.set_xticklabels(
            im.get_xticklabels(),
            fontdict={"fontsize": 14, "fontweight": "bold"},
            rotation=90,
        )
        im.set_yticklabels(
            im.get_yticklabels(),
            fontdict={"fontsize": 14, "fontweight": "bold"},
            rotation=0,
        )

    if not fig_flag:

        fig.subplots_adjust(
            bottom=0.1,
            top=0.85,
            left=0.1,
            right=0.9,
        )

        if not name_lst is None:
            fig.subplots_adjust(wspace=0.5)

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    else:
        plt.show()


def distance2Z(dist_mat, method="ward"):
    # convert the redundant n*n square matrix form into a condensed nC2 array
    distArray = ssd.squareform(dist_mat)
    Z = shc.linkage(distArray, method=method)
    return Z


def most_frequent(List):
    return max(set(List), key=List.count)


def find_measure_color(measure_name, cluster_colors_dict):
    for color in cluster_colors_dict:
        if measure_name in cluster_colors_dict[color]:
            return color
    return "royalblue"


def find_link_colors(
    Z,
    labels,
    cluster_colors_dict,
    extra_colors,
    threshold=None,
):

    extra_colors_copy = extra_colors.copy()
    if threshold is None:
        threshold = 0.7 * np.max(Z[:, 2])

    clstring_labels = shc.fcluster(Z, t=threshold, criterion="distance") - 1
    n_clusters = len(np.unique(clstring_labels))

    cluster_colors = []
    for clstr_id in range(n_clusters):
        # find measures in this cluster
        measures_lst = [labels[i] for i, l in enumerate(clstring_labels) if l == clstr_id]
        # find the color of each measure based on cluster_colors_dict
        colors_lst = [
            find_measure_color(measure_name, cluster_colors_dict)
            for measure_name in measures_lst
        ]
        # find the most frequent color in this cluster
        clstr_color = most_frequent(colors_lst)
        # if the color is not in cluster_colors, add it to cluster_colors
        if clstr_color not in cluster_colors:
            cluster_colors.append(clstr_color)
        # otherwise, add the first color in extra_colors_copy
        else:
            cluster_colors.append(extra_colors_copy[0])
            extra_colors_copy.pop(0)

    # assign each cluster's color to its measures
    cluster_colors_array = [cluster_colors[l] for l in clstring_labels]
    # assign link colors
    link_cols = {}
    for i, i12 in enumerate(Z[:, :2].astype(int)):
        c1, c2 = (link_cols[x] if x > len(Z) else cluster_colors_array[x] for x in i12)
        # the default color is royalblue
        link_cols[i + 1 + len(Z)] = c1 if c1 == c2 else "royalblue"

    return link_cols


def dist_mat_dendo(
    Z,
    labels,
    distances_CI=None,
    threshold=None,
    link_colors=None,
    plot_threshold=False,
    title="",
    save_image=False,
    output_root=None,
):
    """
    distances_CI:
        if  is provided, confidence intervals (CI)
        of the distances will be shown. the order should be the same as Z
    link_colors:
        is a dictionary of colors for each link in Z
        can be used to fix the colors of clusters
    """

    sns.set_context(
        "paper", font_scale=3.5, rc={"lines.linewidth": 3.0, "font.weight": "bold"}
    )

    sns.set_style("darkgrid")

    if threshold is None:
        threshold = 0.7 * np.max(Z[:, 2])

    # link colors
    if link_colors is None:
        link_color_func = None
    else:
        link_color_func = lambda k: link_colors[k]

    width = int(2.5 * len(Z))
    fig = plt.figure(figsize=(width, 5))
    ax = fig.add_subplot(1, 1, 1)
    with mpl.rc_context({"lines.linewidth": 3}):

        dend = shc.dendrogram(
            Z,
            distance_sort="ascending",
            link_color_func=link_color_func,
            no_plot=False,
            labels=labels,
        )

        # show confidence interval of distances
        if not distances_CI is None:

            max_y_lim = None
            for i, d in zip(dend["icoord"], dend["dcoord"]):

                # we have to match the distances in dcoord
                # with those in Z, because the orders are not
                # the same
                count = 0
                for idx, clstr in enumerate(Z):
                    if np.isclose(Z[idx][2], d[1]):
                        count += 1
                        Z_CI = distances_CI[idx]

                if count > 1 or count == 0:
                    warnings.warn("Error in finding std of linkage.", UserWarning)

                x = 0.5 * sum(i[1:3])
                y = d[1]
                ci_line_y = np.linspace(y - Z_CI, y + Z_CI, 100)
                ci_line_x = x * np.ones(100)
                # cut start and the end for better
                # visualization
                ci_line_y = ci_line_y[5:-5]
                ci_line_x = ci_line_x[5:-5]

                plt.plot(ci_line_x, ci_line_y, "black")
                plt.plot(x, y - Z_CI, "k_", markersize=15, linewidth=15)
                plt.plot(x, y + Z_CI, "k_", markersize=15, linewidth=15)
                plt.plot(x, y, "wo", markersize=5, mec="k")
                # plt.annotate("%.2g" % y, (x, y), xytext=(15, 13),
                #             fontsize = 11,
                #             fontweight= 'bold',
                #             textcoords='offset points',
                #             va='top', ha='center')
                if max_y_lim is None:
                    max_y_lim = y + Z_CI
                else:
                    max_y_lim = max(y + Z_CI, max_y_lim)
            plt.ylim(0, max_y_lim * 1.1)

    if plot_threshold:
        plt.axhline(threshold, color="k", linestyle="dashed", linewidth=2)

    if show_title:
        plt.title(title, fontsize=15)

    # set font size of the tick labels and make them bold
    ax.tick_params(axis="x", which="major", labelsize=15)
    ax.tick_params(axis="y", which="major", labelsize=15)
    tick_labels = ax.get_xticklabels() + ax.get_yticklabels()
    for label in tick_labels:
        label.set_fontweight("bold")

    # save figure
    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()


def plot_TSNE(
    dist_mat,
    sample_measure_lst,
    color_dict,
    projection="2d",
    title="",
    save_image=False,
    output_root=None,
):

    sns.set_context(
        "paper", font_scale=2.5, rc={"lines.linewidth": 3.0, "lines.markersize": 10.0}
    )

    sns.set_style("darkgrid")

    fig_width = 20
    fig_height = 20

    if projection == "2d":
        X_embedded = TSNE(
            n_components=2,
            learning_rate="auto",
            init="random",
            perplexity=30,
            metric="precomputed",
        ).fit_transform(dist_mat)

        # 2D plot
        plt.figure(figsize=(fig_width, fig_height))
        sns.scatterplot(
            x=X_embedded[:, 0],
            y=X_embedded[:, 1],
            hue=sample_measure_lst,
            palette=color_dict,
            alpha=0.7,
        )
    elif projection == "3d":
        X_embedded = TSNE(
            n_components=3,
            learning_rate="auto",
            init="random",
            perplexity=30,
            metric="precomputed",
        ).fit_transform(dist_mat)

        measures_lst = list(set(sample_measure_lst))
        measures_lst.sort()

        # 3D plot
        fig = plt.figure(figsize=(fig_width, fig_height))
        ax = fig.add_subplot(projection="3d")
        sample_measure_array = np.array(sample_measure_lst)
        for measure in measures_lst:
            ax.scatter(
                X_embedded[sample_measure_array == measure, 0],
                X_embedded[sample_measure_array == measure, 1],
                X_embedded[sample_measure_array == measure, 2],
                c=color_dict[measure],
                label=measure,
            )
        ax.legend()

    if show_title:
        plt.title(title, fontsize=15)

    # save figure
    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()


def plot_brain_act(act_vec, locs, axes, title="", save_image=False, output_root=""):

    plot_markers(
        node_values=act_vec,
        node_coords=locs,
        node_cmap="hot",
        display_mode="z",
        colorbar=False,
        axes=axes,
    )

    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()


def visualize_state_TC(
    TC_lst, TRs, state_lst, TC_name_lst, title="", save_image=None, output_root=None
):

    color_lst = ["k", "b", "g", "r"]

    if "on" in state_lst and "off" in state_lst:
        ticks = range(2)
    else:
        ticks = range(1, len(state_lst) + 1)

    plt.figure(figsize=(25, 5))
    for i, TC in enumerate(TC_lst):
        plt.plot(TRs, TC, color_lst[i], linewidth=2)
    plt.xlabel("TR")
    plt.yticks(ticks=ticks, labels=state_lst)
    plt.legend(TC_name_lst)
    if show_title:
        plt.title(title)
    if save_image:
        folder = output_root[: output_root.rfind("/")]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(
            output_root + title2file_name(title) + "." + save_fig_format,
            dpi=fig_dpi,
            bbox_inches=fig_bbox_inches,
            pad_inches=fig_pad,
            format=save_fig_format,
        )
        plt.close()
    # else:
    #     plt.show()

    return


################################# Analytical Functions ####################################

from scipy import stats

############## STAT Functions ##############


def make_sim_distribution(sim_mats_lst, name_lst, zip_names=True):
    """
    each sim_mat in the sim_mat_lst corresponds
    to a subj. the name_lst must correspond to the
    columns and rows of sim_mats
    """
    output = {}
    for sim_mat in sim_mats_lst:

        for i, name_i in enumerate(name_lst):
            for j, name_j in enumerate(name_lst):

                if j >= i:
                    continue

                if zip_names:
                    name_i_used = zip_name(name_i)
                    name_j_used = zip_name(name_j)
                else:
                    name_i_used = name_i
                    name_j_used = name_j

                if not name_i_used in output:
                    output[name_i_used] = {}
                if not name_j_used in output[name_i_used]:
                    output[name_i_used][name_j_used] = {"sim": list(), "": list()}

                output[name_i_used][name_j_used]["sim"].append(sim_mat[i, j])
                output[name_i_used][name_j_used][""].append("sim")
    return output


def two_way_anova(data):
    """
    perform two-way anova
    target: sim
    factor1: session
    factor2: direction
    """
    df = pd.DataFrame(data)

    # Performing two-way ANOVA
    model = ols(
        "sim ~ C(session) + C(direction) +\
    C(session):C(direction)",
        data=df,
    ).fit()

    return sm.stats.anova_lm(model, type=2)


def convert_pvalue_to_asterisks(pvalue):
    if pvalue <= 0.0001:
        return "****"
    elif pvalue <= 0.001:
        return "***"
    elif pvalue <= 0.01:
        return "**"
    elif pvalue <= 0.05:
        return "*"
    return "ns"


############## Randomization Functions ##############


def randomize_time(dFC_dict, N):
    """ """
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(dFC_dict):

            dFC_mat_i = dFC_dict[measure_i_name]

            # randomize the temporal order
            n_time = dFC_mat_i.shape[0]
            idx = np.random.choice(n_time, n_time, replace=False)
            dFC_mat_i = dFC_mat_i[idx, :, :]

            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(dFC_dict):

                if j > i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                dFC_mat_j = dFC_dict[measure_j_name]

                # randomize the temporal order
                n_time = dFC_mat_j.shape[0]
                idx = np.random.choice(n_time, n_time, replace=False)
                dFC_mat_j = dFC_mat_j[idx, :, :]

                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]["sim"].append(sim)
                output[measure_i_name][measure_j_name][""].append("sim")

    return output


def suffle_dFC(dFC_mat, mode):
    """
    dFC_mat = ndarray(time, region, region)
    mode can be 'temporal', 'spatial',
    or 'all'
    """
    new_dFC_mat = deepcopy(dFC_mat)
    if mode == "temporal":
        n_time = new_dFC_mat.shape[0]
        new_order = np.random.choice(n_time, n_time, replace=False)
        new_dFC_mat = new_dFC_mat[new_order, :, :]
    elif mode == "spatial":
        n_region = new_dFC_mat.shape[1]
        new_order = np.random.choice(n_region, n_region, replace=False)
        for k, mat in enumerate(new_dFC_mat):
            new_dFC_mat[k, :, :] = mat_reorder(new_dFC_mat[k, :, :], new_order)
    elif mode == "all":
        # spatial
        n_region = new_dFC_mat.shape[1]
        new_order_regions = np.random.choice(n_region, n_region, replace=False)
        for k, mat in enumerate(new_dFC_mat):
            new_dFC_mat[k, :, :] = mat_reorder(new_dFC_mat[k, :, :], new_order_regions)
        # temporal
        n_time = new_dFC_mat.shape[0]
        new_order_time = np.random.choice(n_time, n_time, replace=False)
        new_dFC_mat = new_dFC_mat[new_order_time, :, :]

    return new_dFC_mat


def randomized_dFC_sim(dFC_dict, N, mode):
    """
    mode can be 'temporal', 'spatial',
    or 'all'
    'spatial': this will result in different methods having
    different spatial/region orders but still the
    same temporal order
    'temporal': this will result in different methods having
    different temporal orders but still the
    same spatial order
    'all': this will result in different methods having
    different temporal orders AND different spatial order
    """
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(dFC_dict):

            dFC_mat_i = dFC_dict[measure_i_name]

            # randomize the spatial (regions) order
            dFC_mat_i = suffle_dFC(dFC_mat_i, mode=mode)

            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(dFC_dict):

                if j > i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                dFC_mat_j = dFC_dict[measure_j_name]

                # randomize the temporal order
                dFC_mat_j = suffle_dFC(dFC_mat_j, mode=mode)

                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]["sim"].append(sim)
                output[measure_i_name][measure_j_name][""].append("sim")

    return output


def dFC_rand_generator(FCS, n_time):
    """
    generate a dFC mat of length n_time
    using spatial FC patterns in FCS = (num_pattern, ROI, ROI)
    """
    dFC_rand = None
    idx = np.random.choice(FCS.shape[0], n_time, replace=True)
    dFC_rand = FCS[idx, :, :]
    return dFC_rand


def dFC_rand_sim(FCS_dict, n_time, N):
    """
    for random state TC similarity assessment
    """
    output = {}
    for n in range(N):

        for i, measure_i_name in enumerate(FCS_dict):

            dFC_rand = dFC_rand_generator(FCS_dict[measure_i_name], n_time=n_time)
            dFC_mat_i = dFC_rand
            dFC_mat_i_vec = dFC_mat2vec(dFC_mat_i)

            for j, measure_j_name in enumerate(FCS_dict):

                if j > i:
                    continue
                if not measure_i_name in output:
                    output[measure_i_name] = {}
                if not measure_j_name in output[measure_i_name]:
                    output[measure_i_name][measure_j_name] = {"sim": list(), "": list()}

                dFC_rand = dFC_rand_generator(FCS_dict[measure_j_name], n_time=n_time)
                dFC_mat_j = dFC_rand
                dFC_mat_j_vec = dFC_mat2vec(dFC_mat_j)

                sim, p = stats.spearmanr(dFC_mat_i_vec.flatten(), dFC_mat_j_vec.flatten())
                output[measure_i_name][measure_j_name]["sim"].append(sim)
                output[measure_i_name][measure_j_name][""].append("sim")
    return output


############## Hierarchical Clustering ##############


def correct_order(s):
    list = s.split("-")
    list = [int(item) for item in list]
    list.sort()
    return "-".join(str(x) for x in list)


def open_trees(Z, num_leaf):
    """
    replace trees in Z by their leaves
    """
    Z_copy = deepcopy(Z)
    Z_new = []
    for tree in Z_copy:
        Z_new.append([tree[0], tree[1], tree[2], tree[3]])
    encode_dict = {}
    counter = num_leaf
    for tree in Z_new:
        if tree[0] >= num_leaf:
            tree[0] = encode_dict[tree[0]]
        else:
            tree[0] = str(int(tree[0]))
        if tree[1] >= num_leaf:
            tree[1] = encode_dict[tree[1]]
        else:
            tree[1] = str(int(tree[1]))
        encode_dict[counter] = tree[0] + "-" + tree[1]
        encode_dict[counter] = correct_order(encode_dict[counter])
        counter += 1
    return Z_new


def is_trees_equal(trees_1, trees_2):
    """
    trees_2 is the reference
    """
    for tree in trees_1:
        if (not [tree[0], tree[1]] in trees_2) and (not [tree[1], tree[0]] in trees_2):
            return False
    return True


def is_in_Z_clstrs(trees, Z_clstrs, trees_key):
    for key in Z_clstrs:
        if is_trees_equal(trees, Z_clstrs[key][trees_key]):
            return key
    return None


def cluster_Z(Z_lst, num_leaf):
    """
    Z_lst is the list of linkages of samples
    num_leaf is the number of objects in clustering
    """
    Z_clstrs = {}
    counter = 0
    for Z in Z_lst:
        # replace trees in Z by their leaves
        Z_open = open_trees(Z, num_leaf)
        trees = [[tree[0], tree[1]] for tree in Z_open]
        distances = [tree[2] for tree in Z]
        clstr_idx = is_in_Z_clstrs(trees, Z_clstrs, trees_key="trees")

        if clstr_idx is None:
            Z_clstrs[counter] = {
                "Z": Z,
                "trees": trees,
                "freq": 1,
                "distance_lst": [distances],
            }
            counter += 1
        else:
            Z_clstrs[clstr_idx]["freq"] += 1
            Z_clstrs[clstr_idx]["distance_lst"].append(distances)
    return Z_clstrs



"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import time

import numpy as np
from scipy import signal
from sklearn.covariance import GraphicalLasso, GraphicalLassoCV

from ..dfc import DFC
from ..time_series import TIME_SERIES
from .base_dfc_method import BaseDFCMethod

################################# Sliding-Window #################################

"""

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:
"""


class SLIDING_WINDOW(BaseDFCMethod):

    def __init__(self, **params):
        self.logs_ = ""
        self.TPM = []
        self.FCS_ = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None
        self.graphical_lasso_alpha_ = None

        self.params_name_lst = [
            "measure_name",
            "is_state_based",
            "sw_method",
            "tapered_window",
            "window_std",
            "W",
            "n_overlap",
            "normalization",
            "num_select_nodes",
            "num_time_point",
            "Fs_ratio",
            "noise_ratio",
            "num_realization",
            "session",
        ]
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None

        self.params["measure_name"] = "SlidingWindow"
        self.params["is_state_based"] = False

        assert (
            self.params["sw_method"] in self.sw_methods_name_lst
        ), "sw_method not recognized."

    @property
    def measure_name(self):
        return self.params["measure_name"]  # + '_' + self.sw_method

    def shan_entropy(self, c):
        c_normalized = c / float(np.sum(c))
        c_normalized = c_normalized[np.nonzero(c_normalized)]
        H = -sum(c_normalized * np.log2(c_normalized))
        return H

    def calc_MI(self, X, Y):

        bins = 20

        c_XY = np.histogram2d(X, Y, bins)[0]
        c_X = np.histogram(X, bins)[0]
        c_Y = np.histogram(Y, bins)[0]

        H_X = self.shan_entropy(c_X)
        H_Y = self.shan_entropy(c_Y)
        H_XY = self.shan_entropy(c_XY)

        MI = H_X + H_Y - H_XY
        return MI

    def FC(self, time_series):
        # Graphical Lasso
        if self.params["sw_method"] == "GraphLasso":
            # Standardize the data (zero mean, unit variance for each feature)
            mean = np.mean(time_series, axis=1, keepdims=True)
            std = np.std(time_series, axis=1, keepdims=True)
            time_series_standardized = np.where(std != 0, (time_series - mean) / std, 0)
            model = GraphicalLasso(alpha=self.graphical_lasso_alpha_)
            model.fit(time_series_standardized.T)
            # the covariance matrix will equal the correlation matrix
            C = model.covariance_

        # Mutual information
        elif self.params["sw_method"] == "MI":
            C = np.zeros((time_series.shape[0], time_series.shape[0]))

            for i in range(time_series.shape[0]):
                for j in range(i, time_series.shape[0]):
                    X = time_series[i, :]
                    Y = time_series[j, :]
                    C[j, i] = self.calc_MI(X, Y)

        # Pearson correlation
        else:
            C = np.corrcoef(time_series)
            C[np.isnan(C)] = 0
            # make the diagonal elements 1 (for nan values on the diagonal)
            C[np.diag_indices_from(C)] = 1
        return C

    def dFC(
        self, time_series, W=None, n_overlap=None, tapered_window=False, window_std=None
    ):
        # W is in time samples

        L = time_series.shape[1]
        step = int((1 - n_overlap) * W)
        if step == 0:
            step = 1

        # find the L1 penalty for GraphLasso
        if self.params["sw_method"] == "GraphLasso":
            model = GraphicalLassoCV()
            model.fit(time_series.T)
            self.graphical_lasso_alpha_ = model.alpha_

        FCSs = list()
        TR_array = list()
        for l in range(0, L - W + 1, step):

            # Create rectangular window
            window = np.zeros((L))
            window[l : l + W] = 1

            # Taper the window
            if tapered_window:
                std = window_std if window_std is not None else 3 * W / 22
                window_taper = signal.windows.gaussian(W, std=std)
                window = signal.convolve(window, window_taper, mode="same") / sum(
                    window_taper
                )

            window = np.repeat(
                np.expand_dims(window, axis=0), time_series.shape[0], axis=0
            )
            FCSs.append(self.FC(np.multiply(time_series, window)[:, l : l + W]))
            TR_array.append(int((l + (l + W)) / 2))

        return np.array(FCSs), np.array(TR_array)

    def estimate_FCS(self, time_series):

        return self

    def estimate_dFC(self, time_series):
        """
        we assume calc is applied on subjects separately
        """
        assert (
            len(time_series.subj_id_lst) == 1
        ), "this function takes only one subject as input."

        assert (
            type(time_series) is TIME_SERIES
        ), "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        # W is converted from sec to samples
        FCSs, TR_array = self.dFC(
            time_series=time_series.data,
            W=int(self.params["W"] * time_series.Fs),
            n_overlap=self.params["n_overlap"],
            tapered_window=self.params["tapered_window"],
            window_std=self.params["window_std"],
        )

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=FCSs, TR_array=TR_array, TS_info=time_series.info_dict)

        return dFC


################################################################################
